/*
 *  PalmVNC Version 2.0 - A HiRes VNC client for the Palm OS
 *
 *  Copyright (C) 2003 George Styles
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 *
 */


#include <PalmOS.h>

#include "CollapseUtils.h"
#include "Util.h"

#include "callback.h"

#include "palmVNC.h"
#include "palmVNC_res.h"

#include "KeyMapEdit.h"

#include "KeyMapList.h"

#include "FormUtils.h"

#include "rfbproto.h"
// Used to indicate that there is not index - unlikely to define 9999 key mappings!
#define INDEX_NONE 9999

/******************************************************************
 * exported variables
 *
 *****************************************************************/

DmOpenRef keyMapDbRef;
FormPtr keyMapListForm;

/******************************************************************
 * private variables
 *
 *****************************************************************/
int keyMapMemoryListCount;
KeyMapDbItem keyMapMemoryList[MAX_MEMORY_KEYMAP_COUNT];

/******************************************************************
 * KeyCodesMatch
 * 
 * Returns true if chr and keyCode are both equal.
 *****************************************************************/
static Bool KeyCodesMatch(KeyCodeType aCode, KeyCodeType bCode)
{
	return (aCode.chr == bCode.chr) && (aCode.keyCode == bCode.keyCode);
}



/******************************************************************
 * GetKeyCodeMapAction
 * 
 * Returns the action associated with the given keycode. If none,
 * returns NULL
 *****************************************************************/
static char * GetKeyCodeMapAction(KeyCodeType aCode)
{
	int i;
	
	for (i=0; i<keyMapMemoryListCount; i++)	{
		if (KeyCodesMatch(keyMapMemoryList[i].keyCode,aCode))
			return keyMapMemoryList[i].action;
	}
	return NULL;
	
}


/******************************************************************
 * RunKeyMapAction
 * 
 * Taps the keys for the action assoicated with the given code.
 * 
 *****************************************************************/
Bool RunKeyMapAction(KeyCodeType aCode)
{
	char *action;
	Bool found = false;

	action = GetKeyCodeMapAction(aCode);

	if (action != NULL)	{
		RunActionSequence(action);
		found = true;
	}
	
	return found;
}



/******************************************************************
 * BuildKeymapList
 * 
 * Builds the in-memory list of keymaps - held in memory for rapid searching
 *****************************************************************/
static void BuildInMemoryKeyMapList()
{
	UInt16 index;
	VoidHand recHandle;
	KeyMapDbItem* recPointer;
		
	/* Look for stored KeyMap configurations */
	keyMapMemoryListCount = DmNumRecordsInCategory(keyMapDbRef, dmAllCategories);
	if (keyMapMemoryListCount == 0) {
		return;
	}

	/* Build an index of available keymaps */
	index = 0;
	while ((recHandle = DmQueryNextInCategory(keyMapDbRef, &index, dmAllCategories)) != NULL) {
		recPointer = MemHandleLock(recHandle);
		MemMove(&keyMapMemoryList[index], recPointer, sizeof(KeyMapDbItem));
		MemHandleUnlock(recHandle);
		index++;
	}

}


/******************************************************************
 * KeyMapListInit
 *
 *****************************************************************/
static Boolean KeyMapListInit(Boolean do_populate) {
	ListPtr listPtr;
	static int keymap_list_count = 0;
	static MemHandle keymap_listH = NULL;
	static MemHandle keymapH_H = NULL;
	static MemHandle* keymapH = NULL;
	static Char** keymap_list = NULL;
	char nameBuf[30];

	UInt16 numKeymaps;
	VoidHand recHandle;
	KeyMapDbItem* recPointer;
	UInt16 index;

	/* Clear any old lists */
	listPtr = FrmGetObjectPtr(keyMapListForm, FrmGetObjectIndex(keyMapListForm, KeyMapListKeyMapsList));
	LstSetListChoices(listPtr, NULL, 0);
	
	if (keymap_list_count != 0) {
		for (index = 0; index < keymap_list_count; index++) {
			MemHandleUnlock(keymapH[index]);
			MemHandleFree(keymapH[index]);
		}
		MemHandleUnlock(keymapH_H);
		MemHandleFree(keymapH_H);
		MemHandleUnlock(keymap_listH);
		MemHandleFree(keymap_listH);
		keymap_list_count = 0;
	}
	
	if (do_populate == false) {
		return false;
	}
	
	/* Look for stored KeyMap configurations */
	numKeymaps = DmNumRecordsInCategory(keyMapDbRef, dmAllCategories);
	if (numKeymaps == 0) {
		// And redraw the form
		FrmDrawForm(keyMapListForm);			
		return false;
	}

	/* Build an index of available keymaps */
	keymap_list_count = numKeymaps;
	keymap_listH = MemHandleNew(sizeof(char*) * keymap_list_count);
	keymap_list = MemHandleLock(keymap_listH);
	keymapH_H = MemHandleNew(sizeof(MemHandle) * keymap_list_count);
	keymapH = MemHandleLock(keymapH_H);
	
	index = 0;
	while ((recHandle = DmQueryNextInCategory(keyMapDbRef, &index, dmAllCategories)) != NULL) {
		recPointer = MemHandleLock(recHandle);
	    KeyCodeToKeyName(recPointer->keyCode, nameBuf);		
		keymapH[index] = MemHandleNew(sizeof(Char) * (StrLen(nameBuf) + 1));
		keymap_list[index] = MemHandleLock(keymapH[index]);
		MemMove(keymap_list[index], nameBuf, sizeof(Char) * (StrLen(nameBuf) + 1));
		MemHandleUnlock(recHandle);
		index++;
	}
	
	LstSetListChoices(listPtr, keymap_list, keymap_list_count);
	LstSetSelection(listPtr, 0);

	// And redraw the form
	FrmDrawForm(keyMapListForm);			
	return true;
}

/******************************************************************
 * GetSelectedKeyMapIndex
 *
 * 1. Returns the index of the keymap selected in the list view.
 *****************************************************************/
static UInt16 GetSelectedKeyMapIndex()
{
	ListPtr listPtr;
	listPtr = FrmGetObjectPtr(keyMapListForm, FrmGetObjectIndex(keyMapListForm, KeyMapListKeyMapsList));
	return LstGetSelection(listPtr);
}


/******************************************************************
 * DeleteDBRecord
 *
 * 1. Deletes record from DB
 *
 *****************************************************************/
static Bool DeleteDBRecord(UInt16 AIndex)
{
	DmRemoveRecord(keyMapDbRef, AIndex);
	return true;
}


/******************************************************************
 * LoadDBRecord
 *
 * 1. Loads the specified record from database
 *
 * Returns true if successful??
 *****************************************************************/
static Bool LoadDBRecord(KeyMapDbItem *AKeyMap, UInt16 AIndex)
{
	VoidHand recHandle;
	VoidPtr recPointer;

	Bool loaded = false;

	// Load an exisiting profile.
	recHandle = DmGetRecord(keyMapDbRef, AIndex);
	if (recHandle == NULL) {
		return false;
	} //if
	
	recPointer = MemHandleLock(recHandle);
	if (recPointer == NULL) {
		return false;
	} // if
	
	// Blank out record before loading
	MemSet(AKeyMap, sizeof(KeyMapDbItem), (Byte)0);

	// Copy data to AKeyMap	
	MemMove(AKeyMap, recPointer, sizeof(KeyMapDbItem));
	
	MemHandleUnlock(recHandle);

	DmReleaseRecord(keyMapDbRef, AIndex, true);

	// Be sure to set AKeyMap->Index to correct value
	AKeyMap->index = AIndex;
	return true;
}

/******************************************************************
 * SaveDBRecord
 *
 * 1. Saves the specified record to the database
 * ASSUMES THAT THE RECORD HAS A VALID index VALUE
 *
 * Returns true if successful??
 *****************************************************************/
static Bool SaveDBRecord(KeyMapDbItem *AKeyMap)
{
	VoidHand recHandle;
	VoidPtr recPointer;

	UInt16 AIndex;

	AIndex = AKeyMap->index;

	// Overwrite an existing profile.
	recHandle = DmGetRecord(keyMapDbRef, AIndex);
	if (recHandle == NULL) {
		return false;
	} //if
			
	recPointer = MemHandleLock(recHandle);
	if (recPointer == NULL) {
		return false;
	} //if
	
	DmWrite(recPointer, 0, AKeyMap, sizeof(KeyMapDbItem));
	MemHandleUnlock(recHandle);
	DmReleaseRecord(keyMapDbRef, AIndex, true);

	return true;
}


/******************************************************************
 * CreateNewDBRecord
 *
 * 1. Create a new DB record
 *
 * Returns new DB index.
 *****************************************************************/
static UInt16 CreateNewDBRecord(KeyMapDbItem *AKeyMap)
{
	VoidHand recHandle;
	VoidPtr recPointer;

	UInt16 newIndex;

	newIndex = dmMaxRecordIndex; // New record
	recHandle = DmNewRecord(keyMapDbRef, &newIndex, sizeof(KeyMapDbItem));
	recPointer = MemHandleLock(recHandle);
	DmWrite(recPointer, 0, AKeyMap, sizeof(KeyMapDbItem));
	MemHandleUnlock(recHandle);
	DmReleaseRecord(keyMapDbRef, newIndex, true);
	AKeyMap->index = newIndex;
	return newIndex;
}

/******************************************************************
 * AddNewKeymap
 *
 * 1. Create a new keymap
 * 2. Popup keymap editor
 * 3. If OK clicked, save new keymap to database.
 *****************************************************************/
static void AddNewKeyMap()
{
	KeyMapDbItem newKeyMap;
	
//	char *dummy_name = "My keymap\0";
	char *dummy_action = "My Action\0";

	// Generate new 'blank' record
	newKeyMap.index = INDEX_NONE;
//	MemMove(newKeyMap.name, dummy_name, StrLen(dummy_name)+1);
	newKeyMap.keyCode.chr = 0;
	newKeyMap.keyCode.keyCode = 0;
	newKeyMap.keyCode.modifiers = 0;
	MemMove(newKeyMap.action, dummy_action, StrLen(dummy_action)+1);		
	
	// Popup editor
	if (DoKeyMapEdit(&newKeyMap)) {
	  	// Need to save the new record in database
		CreateNewDBRecord(&newKeyMap);
		// And rebuild and refresh list on screen
		KeyMapListInit(true);
	} //if
} // AddNewKeyMap()

/******************************************************************
 * EditSelectedKeymap
 *
 * 1. Load specified keymap
 * 2. Popup keymap editor
 * 3. If OK clicked, save keymap to database, over existing record
 *****************************************************************/
static void EditSelectedKeyMap()
{
	KeyMapDbItem tempKeyMap;
	UInt16 AIndex = GetSelectedKeyMapIndex();
	
	// Load record from DB
	LoadDBRecord(&tempKeyMap, AIndex);
		
	// Popup editor
	if (DoKeyMapEdit(&tempKeyMap)) {
	  	// Need to save record over old one in database
		SaveDBRecord(&tempKeyMap);
		// And refresh list on screen
		KeyMapListInit(true);
	} //if
} // EditExistingKeyMap()

/******************************************************************
 * DeleteSelectedKeymap
 *
 * 1. Delete selected keymap.
 *****************************************************************/
static void DeleteSelectedKeyMap()
{
	UInt16 AIndex = GetSelectedKeyMapIndex();
	
	// Load record from DB
	DeleteDBRecord(AIndex);
		
	// And refresh list on screen
	KeyMapListInit(true);
} // DeleteSelectedKeymap


/******************************************************************
 * OpenDatabase
 *
 * Opens the database, or converts database from previous version, and load preferences
 *****************************************************************/
DmOpenRef KeyMapOpenDatabase() {
  LocalID dbID;
//  DmOpenRef dbRef;
  UInt16 attributes;
  Err err;
	
	
  dbID = DmFindDatabase(0, KEYMAP_DB_NAME);
  
  if (! dbID) {
  	// DB not found, create it
  	err = DmCreateDatabase(0, KEYMAP_DB_NAME, CREATOR_ID, KEYMAP_DB_TYPE, false);
  	if (err) {
  		return NULL;
  	}

    dbID = DmFindDatabase(0, KEYMAP_DB_NAME);

		// set the backup bit
		DmDatabaseInfo( 0, dbID, NULL, &attributes, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL );
		attributes |= dmHdrAttrBackup;
		DmSetDatabaseInfo( 0, dbID, NULL, &attributes, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL );
  }
  
  // open the DB
  keyMapDbRef = DmOpenDatabase(0, dbID, dmModeReadWrite);
   
  // Read it into memory list of keymaps
  BuildInMemoryKeyMapList();
  return keyMapDbRef;
}

/******************************************************************
 * CloseDatabase
 *
 * Save preferences information.
 *****************************************************************/
void KeyMapCloseDatabase(DmOpenRef db) {
  DmCloseDatabase(db);
}


/******************************************************************
 * KeyMapListFormHandleEvent
 *
 * Handles events in the KeyMapList selection form
 *****************************************************************/
static Boolean KeyMapListFormHandleEvent(EventPtr e) {
	Boolean handled, closeKeyMapListForm;
	FormPtr frm;
//	UInt16 index;
//	ListPtr listPtr;

	CALLBACK_PROLOGUE

	frm = FrmGetActiveForm();
	handled = false;
	closeKeyMapListForm = false;

	switch (e->eType) {
		case frmOpenEvent:
  		CollapseSetState(frm, collapseStateUser);
			FrmDrawForm(frm);
			break;

		case frmCloseEvent:
			closeKeyMapListForm = true;
			handled = true;
			break;

		case ctlSelectEvent:
			switch(e->data.ctlSelect.controlID) {
				case KeyMapListNewButton:
					AddNewKeyMap();
					handled = true;
					break;
				case KeyMapListEditButton:
					EditSelectedKeyMap();
					handled = true;
					break;
				case KeyMapListDeleteButton:
					DeleteSelectedKeyMap();
					handled = true;
					break;
				default:
					break;
			}
			break;

		default:
			break;
	}

	if (closeKeyMapListForm) {
		// the memory under field text handles is freed by the UI manager
		CALLBACK_EPILOGUE
		FrmReturnToForm(MainForm);
	}

	CALLBACK_EPILOGUE
	return handled;
}


/******************************************************************
 * KeyMapList
 *
 * Generates the KeyMapList dialog
 *****************************************************************/
void DoKeyMapList() {
	FormPtr previousForm = FrmGetActiveForm();
	UInt16 hitButton;
	
	// Get form
	keyMapListForm = FrmInitForm(KeyMapListForm);	
	
	FrmSetActiveForm(keyMapListForm);
	FrmSetEventHandler(keyMapListForm, KeyMapListFormHandleEvent);
	
	// Read in list of keymaps from database
	KeyMapListInit(true); // true = populate after clearing

	hitButton = FrmDoDialog(keyMapListForm);
	
	/* Empty list */
	KeyMapListInit(false);

	if (previousForm) {
		FrmSetActiveForm(previousForm);
	}
	FrmDeleteForm(keyMapListForm);
	
	// And re-read list of keymaps for memory list
	BuildInMemoryKeyMapList();
}



