/*
 *  PalmVNC Version 2.0 - A HiRes VNC client for the Palm OS
 *
 *  Copyright (C) 2003 Olivier Grardin (olivier@palmattitude.org)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 *
 */


#include <PalmOS.h>

#include "CollapseUtils.h"
#include "Util.h"

#include "callback.h"

#include "palmVNC.h"
#include "palmVNC_res.h"

#include "MacroEdit.h"

#include "MacroList.h"



/******************************************************************
 * exported variables
 *
 *****************************************************************/

DmOpenRef macrosDbRef;


/******************************************************************
 * OpenDatabase
 *
 * Opens the database, or converts database from previous version, and load preferences
 *****************************************************************/
DmOpenRef MacroOpenDatabase() {
  LocalID dbID;
  DmOpenRef dbRef;
  UInt16 attributes;
  Err err;
	
	
  dbID = DmFindDatabase(0, MACROS_DB_NAME);
  
  if (! dbID) {
  	// DB not found, create it
  	err = DmCreateDatabase(0, MACROS_DB_NAME, CREATOR_ID, MACROS_DB_TYPE, false);
  	if (err) {
  		return NULL;
  	}

    dbID = DmFindDatabase(0, MACROS_DB_NAME);

		// set the backup bit
		DmDatabaseInfo( 0, dbID, NULL, &attributes, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL );
		attributes |= dmHdrAttrBackup;
		DmSetDatabaseInfo( 0, dbID, NULL, &attributes, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL );
  }
  
  // open the DB
  dbRef = DmOpenDatabase(0, dbID, dmModeReadWrite);
   
  return dbRef;
}

/******************************************************************
 * CloseDatabase
 *
 * Save preferences information.
 *****************************************************************/
void MacroCloseDatabase(DmOpenRef db) {
  DmCloseDatabase(db);
}

/******************************************************************
 * MacroListInit
 *
 *****************************************************************/
static Boolean MacroListInit(FormType * frm, Boolean do_populate) {
	ListPtr listPtr;
	static int macros_list_count = 0;
	static MemHandle macros_listH = NULL;
	static MemHandle macroH_H = NULL;
	static MemHandle* macroH = NULL;
	static Char** macros_list = NULL;

	UInt16 numMacros;
	VoidHand recHandle;
	MacrosDbItem* recPointer;
	UInt16 index;

	/* Clear any old lists */
	listPtr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, MacroListMacrosList));
	LstSetListChoices(listPtr, NULL, 0);
	
	if (macros_list_count != 0) {
		for (index = 0; index < macros_list_count; index++) {
			MemHandleUnlock(macroH[index]);
			MemHandleFree(macroH[index]);
		}
		MemHandleUnlock(macroH_H);
		MemHandleFree(macroH_H);
		MemHandleUnlock(macros_listH);
		MemHandleFree(macros_listH);
		macros_list_count = 0;
	}
	
	if (do_populate == false) {
		return false;
	}
	
	/* Look for stored Macro configurations */
	numMacros = DmNumRecordsInCategory(macrosDbRef, dmAllCategories);
	if (numMacros == 0) {
		return false;
	}

	/* Build an index of available macros */
	macros_list_count = numMacros;
	macros_listH = MemHandleNew(sizeof(char*) * macros_list_count);
	macros_list = MemHandleLock(macros_listH);
	macroH_H = MemHandleNew(sizeof(MemHandle) * macros_list_count);
	macroH = MemHandleLock(macroH_H);
	
	index = 0;
	while ((recHandle = DmQueryNextInCategory(macrosDbRef, &index, dmAllCategories)) != NULL) {
		recPointer = MemHandleLock(recHandle);
		macroH[index] = MemHandleNew(sizeof(Char) * (StrLen(recPointer->name) + 1));
		macros_list[index] = MemHandleLock(macroH[index]);
		MemMove(macros_list[index], recPointer->name, sizeof(Char) * (StrLen(recPointer->name) + 1));
		MemHandleUnlock(recHandle);
		index++;
	}
	
	LstSetListChoices(listPtr, macros_list, macros_list_count);
	LstSetSelection(listPtr, 0);
		
	return true;
}


/******************************************************************
 * LoadMacro
 *
 * Copies the macro settings from the appropriate database record
 *****************************************************************/
static void LoadMacro(DmOpenRef db, MacrosDbItem *currentMacro, UInt16 index)
{
	VoidHand recHandle;
	VoidPtr recPointer;

	char *dummy_name = "My macro\0";

	Bool loaded = false;

	if (index != dmMaxRecordIndex) {
		// Load an exisiting profile.
		recHandle = DmGetRecord(db, index);
		if (recHandle != NULL) {
			recPointer = MemHandleLock(recHandle);
			if (recPointer != NULL) {
				MemSet(currentMacro, sizeof(MacrosDbItem), (Byte)0);
				MemMove(currentMacro, recPointer, sizeof(MacrosDbItem));
				loaded = true;
				currentMacro->index = index;
				MemHandleUnlock(recHandle);
			}
			DmReleaseRecord(db, index, true);
		}
	}

	if (loaded == false) {
		// fill default values
		currentMacro->index = 0;
		MemMove(currentMacro->name, dummy_name, StrLen(dummy_name)+1);
		currentMacro->shift = false;
		currentMacro->ctrl = false;
		currentMacro->alt = false;
	}

	return;
}

/******************************************************************
 * SaveMacro
 *
 *  Copies the current Macro settings to the database
 *****************************************************************/
static UInt16 SaveMacro(DmOpenRef db, MacrosDbItem *currentMacro)
{
	VoidHand recHandle;
	VoidPtr recPointer;

	UInt16 local_index;

	local_index = currentMacro->index;

	if (local_index > 0) {
		// Overwrite an existing profile.
		recHandle = DmGetRecord(db, currentMacro->index);
		if (recHandle != NULL) {
			recPointer = MemHandleLock(recHandle);
			if (recPointer == NULL) {
				local_index = 0;
			}
			else {
				MemHandleUnlock(recHandle);
			}
			DmReleaseRecord(db, currentMacro->index, true);
		}
		else {
			local_index = 0;
		}
	}

	if (local_index == 0) {
		// Create a new profile.
		local_index = dmMaxRecordIndex;
		recHandle = DmNewRecord(db, &local_index, sizeof(MacrosDbItem));
		recPointer = MemHandleLock(recHandle);
		DmWrite(recPointer, 0, currentMacro, sizeof(MacrosDbItem));
		MemHandleUnlock(recHandle);
		DmReleaseRecord(db, local_index, true);
	}
	else {
		recHandle = DmGetRecord(db, local_index);
		recPointer = MemHandleLock(recHandle);
		DmWrite(recPointer, 0, currentMacro, sizeof(MacrosDbItem));
		MemHandleUnlock(recHandle);
		DmReleaseRecord(db, local_index, true);
	}
	currentMacro->index = local_index;

	return local_index;
}


/******************************************************************
 * DeleteMacro
 *
 *  Removes currentMacro connection settings from the database
 *****************************************************************/
static void DeleteMacro(DmOpenRef db, UInt16 index)
{
	UInt16 local_index;

	local_index = index;

	if (local_index > 0) {
		// Delete an existing profile.
		DmRemoveRecord(db, index);
	}
}





/*
 * DESCRIPTION:
 *
 * This routine is the event handler for the "MacroList" of this 
 * application.
 *
 * PARAMETERS:
 *
 * eventP
 *     a pointer to an EventType structure
 *
 * RETURNED:
 *     true if the event was handled and should not be passed to
 *     FrmHandleEvent
 */

/******************************************************************
 * MacroListFormHandleEvent
 *
 * Handles events in the MacroList selection form
 *****************************************************************/
static Boolean MacroListFormHandleEvent(EventPtr e) {
	Boolean handled, closeMacroListForm;
	FormPtr frm;
	UInt16 index;
	ListPtr listPtr;

	CALLBACK_PROLOGUE

	frm = FrmGetActiveForm();
	handled = false;
	closeMacroListForm = false;

	switch (e->eType) {
		case frmOpenEvent:
  		CollapseSetState(frm, collapseStateUser);
			FrmDrawForm(frm);
			break;

		case frmCloseEvent:
			closeMacroListForm = true;
			handled = true;
			break;

		case ctlSelectEvent:
			switch(e->data.ctlSelect.controlID) {

				case MacroListNewButton:
				case MacroListEditButton:
					if (e->data.ctlSelect.controlID == MacroListNewButton) {
						index = dmMaxRecordIndex;
					}
					else {
						listPtr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, MacroListMacrosList));
						index = LstGetSelection(listPtr);
					}
					LoadMacro(macrosDbRef, &currentMacro, index);
					if (DoMacroEdit()) {
						SaveMacro(macrosDbRef, &currentMacro);
					}
					else {
						LoadMacro(macrosDbRef, &currentMacro, index);
					}
					MacroListInit(frm, true);
					FrmDrawForm(frm);
					handled = true;
					break;

				case MacroListCopyButton:
				case MacroListDeleteButton:
					listPtr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, MacroListMacrosList));

					switch (e->data.ctlSelect.controlID) {
						case MacroListCopyButton:
							index = LstGetSelection(listPtr);
							if (index != noListSelection) {
								LoadMacro(macrosDbRef, &currentMacro, index);
								SaveMacro(macrosDbRef, &currentMacro);
								MacroListInit(frm, true);
								FrmDrawForm(frm);
							}
							handled = true;
							break;

						case MacroListDeleteButton:
							index = LstGetSelection(listPtr);
							if (index != noListSelection) {
								LoadMacro(macrosDbRef, &currentMacro, index);
								if (FrmCustomAlert(DeleteConfirmAlert, currentMacro.name, "", "") == 0) {
									DeleteMacro(macrosDbRef, index);
									MacroListInit(frm, true);
									FrmDrawForm(frm);
								}
							}
							handled = true;
							break;
					}
					break;

				default:
					 break;
			}
			break;

		default:
			break;
	}

	if (closeMacroListForm) {
		// the memory under field text handles is freed by the UI manager
		CALLBACK_EPILOGUE
		FrmReturnToForm(MainForm);
	}

	CALLBACK_EPILOGUE
	return handled;
}


/******************************************************************
 * MacroList
 *
 * Generates the MacroList dialog
 *****************************************************************/
void DoMacroList() {
	FormPtr previousForm = FrmGetActiveForm();
	FormPtr frm = FrmInitForm(MacroListForm);
	UInt16 hitButton;
	
	FrmSetActiveForm(frm);
	FrmSetEventHandler(frm, MacroListFormHandleEvent);
	
	MacroListInit(frm, true);

	hitButton = FrmDoDialog(frm);
	
	/* Empty list */
	MacroListInit(frm, false);

	if (previousForm) {
		FrmSetActiveForm(previousForm);
	}
	FrmDeleteForm(frm);
}


