/*
 *  PalmVNC Version 2.0 - A HiRes VNC client for the Palm OS
 *
 *  Copyright (C) 2002 Olivier Grardin (olivier@palmattitude.org)
 *
 *  Version 1.40 Copyright (C) 2001 Harakan Software
 *  Version 1.1  Copyright (C) 1998 International Computer 
 *    Science Institute (ICSI)
 *  Author: Vladimir Minenko, minenko@icsi.berkeley.edu
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 *
 */


#include <PalmOSGLue.h>
#include <PalmOS.h>

// Sony CLIE SDK
#include <SonyCLIE.h>

// Palm 5-way nav
#include "PalmChars.h"

// HandEra QVGA
#include <Vga.h>

#include "CollapseUtils.h"
#include "CommonGraphics.h"

#include "palmVNC.h"
#include "palmVNC_res.h"

#include "sockets.h"

// build number (generated during make process)
#include "build.h"

// form interfaces
#include "PrefsForm.h"
#include "PasswordForm.h"
#include "ServerSelect.h"
#include "ServerEdit.h"
#include "MacroList.h"
#include "KeyMapList.h"
#include "FormUtils.h"
#include "Util.h"

#include "ColorPicker.h"

// VNC keysyms
#define XK_MISCELLANY
#include "keysymdef.h"


#define PALMVNC_MIN_OS     sysMakeROMVersion(3,0,0,sysROMStageRelease,0) // PalmOS 3.0

//#define INVALID_PIXEL       0xffffffff

#define COLORMAP_SIZE       256

// switch defines for scrollbars
// we have our own scrollbars because the native scrollbars are too large
#define HORIZ           1
#define HORIZ_UP        2
#define HORIZ_DOWN      3
#define HORIZ_TOP       4
#define HORIZ_BUTTOM    5

#define VERT            6
#define VERT_UP         7
#define VERT_DOWN       8
#define VERT_TOP        9
#define VERT_BUTTOM     10

#define MAIN_PRESSED    11
#define MAIN_MOVED      12
#define MAIN_ONCE       13
#define MAIN_DOUBLE     14


static unsigned long lastPenTime = 0;
static unsigned short lastX = 0, lastY = 0;
static unsigned short currentMouseButton = 0;

// off-screen panning window
static WinHandle offScreenWindow = 0;

// bits of the off-screen panning window
// (this is only used in legacy 160x160 modes)
static char *offScreenWindowBits = NULL;

// Off-screen keyboard bitmap window
static WinHandle keyboard = 0;

// Define keyboard state bits
static unsigned short keyboard_state = 0;

#define KEYBOARD_VISIBLE 7
#define KEYBOARD_ESC 0
#define KEYBOARD_SHIFT 1
#define KEYBOARD_CTRL 2
#define KEYBOARD_ALT 3
#define KEYBOARD_FN 4

#define KEYBOARD_KEY_WIDTH 20
#define KEYBOARD_KEY_COUNT 5

static Coord keyboardY = 150;
static Coord keyboardWidth, keyboardHeight;


// on-screen main window
static WinHandle topLevel = 0;

// blocks screen updating when true (e.g. when the menu is displayed)
static Boolean dontDraw = false;
#define DrawingBlocked() dontDraw
	/* Was previously handled the following way:
	
					if (MenuGetActiveMenu())
						return (MenuGetActiveMenu()->attr.visible);
			
		This is not supported any more in PalmOS 5, because the "visible"
		attribute is not exposed.
		
		Now handled by watching winEnterEvent and winExitEvent,
		and setting the dontDraw flag, conform to PalmSource 
		recommendations. Functions changed to a macro for speed.
	*/


// Screen mode
static UInt32 screenDepth = 0; 	// 2: 4 greys, 4: 16 greys, 8: 256 colors (palette)
																// 16: direct color
static UInt32 screenWidth = 0;
static UInt32 screenHeight = 0;
static int density = 1;

// reference number of Sony HiRes Library
static UInt16 sonyHiResRefNum = sysInvalidRefNum;

static Char palmvncVersion[40];

typedef UInt8 BGR233ColorType;

// BGR233 codes for minimal greyscale colours
#define BLACK       0x00
#define DARK_GRAY   0x52
#define LIGHT_GRAY  0xA4
#define WHITE       0xFF
#define DARK_RED    0x04
#define DARK_BLUE   0x80

static IndexedColorType BGR233ToIntensityNibble[COLORMAP_SIZE] = {
 0, 1, 3, 4, 5, 7, 8, 9,  1, 2, 3, 5, 6, 7, 9,10,  1, 3, 4, 5, 7, 8, 9,11,  2, 3, 5, 6, 7, 9,10,12,
 3, 4, 5, 7, 8,10,11,12,  3, 5, 6, 7, 9,10,12,13,  4, 5, 7, 8,10,11,12,14,  5, 6, 8, 9,10,12,13,14,
 1, 2, 3, 5, 6, 7, 9,10,  1, 3, 4, 5, 7, 8, 9,11,  2, 3, 5, 6, 7, 9,10,11,  3, 4, 5, 7, 8, 9,11,12,
 3, 5, 6, 7, 9,10,11,13,  4, 5, 7, 8, 9,11,12,13,  5, 6, 7, 9,10,11,13,14,  5, 7, 8, 9,11,12,13,15,
 1, 2, 4, 5, 6, 8, 9,10,  2, 3, 4, 6, 7, 8,10,11,  2, 4, 5, 6, 8, 9,11,12,  3, 4, 6, 7, 9,10,11,13,
 4, 5, 6, 8, 9,11,12,13,  4, 6, 7, 9,10,11,13,14,  5, 7, 8, 9,11,12,13,15,  6, 7, 9,10,11,13,14,15,
 2, 3, 4, 6, 7, 8,10,11,  2, 4, 5, 6, 8, 9,10,12,  3, 4, 6, 7, 8,10,11,12,  4, 5, 6, 8, 9,10,12,13,
 4, 6, 7, 8,10,11,12,14,  5, 6, 8, 9,10,12,13,14,  6, 7, 8,10,11,12,14,15,  6, 8, 9,10,12,13,14,15
};


// the screen API to be used
ScreenMode VNC_screenMode;

static Char* SCREEN_MODE_DESC[SCREEN_MODE_COUNT] =
{
	"(Undefined)",
	"OS 5 Platform API",
	"Sony HiRes API",
	"Legacy 8bpp (256 colors)",
	"Legacy 4bpp (16 greys)",
	"Legacy 2bpp (4 greys)",
	"HandEra 330 QVGA",
};


ServersDbItem currentServer;
MacrosDbItem currentMacro;

VncPreferenceType VNC_prefs;


// pen event processing
static int downX, downY;
static unsigned short buttonState = 0;
static int viewportX = 0, viewportY = 0;

static Boolean pen_zooming = false;

static RectangleType vertScrollbarRect;
static RectangleType horizScrollbarRect;
static RectangleType vertSlideRect;
static RectangleType horizSlideRect;

static RectangleType updateRect;

//static RectangleType virtualBounds;

UInt16 AppNetRefnum;
static Boolean networkIsOpen = false;





/******************************************************************
 * VncSetForeColor
 *
 * Sets the foregroung ink to the specified BGR233 color,
 * regardless of mode.
 *****************************************************************/
static void VncSetForeColor(register BGR233ColorType color) {
	RGBColorType newColor;

	if (screenDepth > 8) {
		// RGB mode, extract RGB values from BGR233ColorType
		newColor.index = 0;
		newColor.r = (color &0x07) << 5;
		newColor.g = (color &0x38) << 2 ;
		newColor.b = (color &0xC0) << 0;
		WinSetForeColorRGB(&newColor, NULL);
		WinSetTextColorRGB(&newColor, NULL);
	}
	else if (screenDepth == 8) {
		// palette mode, IndexedColorType = BGR233ColorType
		WinSetForeColor((IndexedColorType) color);
		WinSetTextColor((IndexedColorType) color);
	}
	else if (VNC_screenMode == modeSonyHires) {
		// greyscale HiRes (Cli T415)
		WinSetForeColor((IndexedColorType) (0x0F ^ BGR233ToIntensityNibble[color]));
	}
	else {
		// B&W modes, convert BGR233ColorType to intensity
		newColor.index = 0;
		newColor.r = (BGR233ToIntensityNibble[color] << 4) | BGR233ToIntensityNibble[color];
		newColor.g = newColor.r;
		newColor.b = newColor.r;
		WinSetColors(&newColor, 0, 0, 0);
	}
}


/***********************************************************************
 *
 * RomVersionCompatible
 *
 * This routine checks that a ROM version meets your
 * minimum requirement and sets the version number in global variable
 ***********************************************************************/
static UInt32 RomVersionCompatible(UInt32 requiredVersion, UInt16 launchFlags)
{
  UInt32 romVersion;

  // See if we have at least the minimum required version of the ROM or later.
  FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);

  if (romVersion < requiredVersion) {
		if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp))
				== (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) {
		  FrmAlert(VersionAlert);

		  // Pilot 1.0 will continuously relaunch this app unless we switch to
		  // another safe one.
		  if (romVersion < sysMakeROMVersion(1,0,0,sysROMStageRelease,0)) {
				AppLaunchWithCommand(sysFileCDefaultApp, sysAppLaunchCmdNormalLaunch, NULL);
		  }
		}
		return sysErrRomIncompatible;
  }

  return 0;
}


/******************************************************************
 * PostProgressMessage
 *
 * writes a message and draws a simple progress bar
 *****************************************************************/
void PostProgressMessage(char *msgStr, unsigned short curr, unsigned short max)
{
	WinHandle prevDrawWin;
  static RectangleType r = {{20, 140}, {120, 10}};
  unsigned short length = StrLen(msgStr);
  
	if (DrawingBlocked()) {
  	return;
  }

  if (max && VNC_prefs.disableProgressBar) {
  	// progress bar disabled in preferences
  	return;
  }

	if (VNC_screenMode == modeHiDensity) {
		WinPushDrawState();
		prevDrawWin = WinSetDrawWindow(WinGetDisplayWindow());
		WinSetCoordinateSystem(kCoordinatesStandard);
	}
	
  VncSetForeColor(DARK_GRAY);

	r.topLeft.y = screenHeight/density - 20;
  WinEraseRectangle(&r, 0);

  WinDrawChars(msgStr, length,
		80 - FntLineWidth(msgStr, length) / 2, r.topLeft.y - 1);

  WinDrawRectangleFrame(simpleFrame, &r);

  if (max) {
		r.extent.x = (120 * curr) / max;
		WinInvertRectangle(&r, 0);
  }

  VncSetForeColor(BLACK);

	if (VNC_screenMode == modeHiDensity) {
		WinSetDrawWindow(prevDrawWin);
		WinPopDrawState();
	}

}


/******************************************************************
 * GetUpdateRectAtPos
 *
 * Re-assigns the coordinates of the update rect. and returns the
 * pointer to it.
 *****************************************************************/
static RectanglePtr GetUpdateRectAtPos(unsigned short x, unsigned short y) {
  updateRect.topLeft.x = x;
  updateRect.topLeft.y = y;
  return &updateRect;
}


/******************************************************************
 * OffscreenDrawFilledRectangle
 *
 * Draws a filled rectangle in the offscreen window If adjust is true
 * x and y coordinates are translated to the origin in the virtual
 *****************************************************************/
void OffscreenDrawFilledRectangle(unsigned short color,
				short x, short y,
				unsigned short width, unsigned short height,
				int adjust)
{
	short row = 0;
	short column = 0;
	short start_column = 0;
	short end_column = 0;
	short start_row = 0;
	short end_row = 0;

	// align to real coordinates
	if (adjust)	{
		x -= palmFB.virtual.x;
		y -= palmFB.virtual.y;
	}

	// Check top-left corner
	if ((x >= palmFB.virtual.w) || (y >= palmFB.virtual.h))	{
		return;
	}

	// Check top and bottom boundaries
	if (y < 0) {
		start_row = -y;
	}
	else {
		start_row = 0;
	}
	if ((y + height) >= palmFB.virtual.h) {
		end_row = (palmFB.virtual.h - y) - 1;
	}
	else {
		end_row = height - 1;
	}
	if (end_row < start_row) {
		return;
	}

	// Check left and right boundaries
	if (x < 0) {
		start_column = -x;
	}
	else {
		start_column = 0;
	}
	if ((x + width) >= palmFB.virtual.w) {
		end_column = (palmFB.virtual.w - x) - 1;
	}
	else {
		end_column = width - 1;
	}
	if (end_column < start_column) {
		return;
	}


	if (VNC_screenMode == modeSonyHires 
			|| VNC_screenMode == modeHiDensity
			|| VNC_screenMode == modeHandEra330) {
		RectangleType rect;
	  WinHandle prevDrawWin = WinSetDrawWindow(offScreenWindow);

		rect.topLeft.x = start_column + x;
		rect.topLeft.y = start_row + y;
		rect.extent.x = end_column - start_column + 1;
		rect.extent.y = end_row - start_row + 1;

	  if (VNC_screenMode == modeHiDensity) {
	  	WinPushDrawState();
	  	WinSetCoordinateSystem(kCoordinatesNative);
	  }

	  VncSetForeColor(color);
		CommonWinDrawRectangle(&rect, 0);
		
	  if (VNC_screenMode == modeHiDensity) {
	  	WinPopDrawState();
	  }
	  
	  WinSetDrawWindow(prevDrawWin);
	}
	else if (screenDepth == 8) {
		/* Do whole region */
		for (row = start_row; row <= end_row; row++) {
			for (column = start_column; column <= end_column; column ++) {
				offScreenWindowBits[(((row + y) * screenWidth) + column + x)] = color;
			}
		}
	}
	else if (screenDepth == 4) {
		color = BGR233ToIntensityNibble[color];
		/* Do left-hand column if non-aligned */
		if (((x+start_column) % 2))	{ 			 /* 1, 3, 5, ...  */
			for (row = start_row; row <= end_row; row++) {
				offScreenWindowBits[(((row + y) * screenWidth) + start_column + x) / 2] |= (0xF);
				offScreenWindowBits[(((row + y) * screenWidth) + start_column + x) / 2] ^= (color);
			}
			start_column++;
		}

		/* Check if only 1 column */
		if (end_column < start_column) {
			return;
		}

		/* Do right-hand column if non-aligned */
		if (!((x+end_column) % 2)) { /* 0, 2, 4, ...  */
			for (row = start_row; row <= end_row; row++) {
				offScreenWindowBits[(((row + y) * screenWidth) + end_column + x) / 2] |= (0xF << 4);
				offScreenWindowBits[(((row + y) * screenWidth) + end_column + x) / 2] ^= (color << 4);
			}
			end_column--;
		}

		/* Check if only 2 columns */
		if (end_column <= start_column) {
			return;
		}

		/* Do middle region */
		color = ~(color + (color << 4));
		for (row = start_row; row <= end_row; row++) {
			for (column = start_column; column < end_column; column += 2)	{
				offScreenWindowBits[(((row + y) * screenWidth) + column + x) / 2] = color;
			}
		}
	}
	else if (screenDepth == 2) {
		/* Very slow implementation for now to sanity-check */

		color = (BGR233ToIntensityNibble[color] >> 2) & 0x3;

		for (row = start_row; row <= end_row; row++) {
			for (column = start_column; column <= end_column; column++)	{
				switch ((x + column) % 4) {
					case 0:
						offScreenWindowBits[(((row + y) * screenWidth) + column + x) / 4] |= (0x3 << 6);
						offScreenWindowBits[(((row + y) * screenWidth) + column + x) / 4] ^= (color << 6);
						break;
					case 1:
						offScreenWindowBits[(((row + y) * screenWidth) + column + x) / 4] |= (0x3 << 4);
						offScreenWindowBits[(((row + y) * screenWidth) + column + x) / 4] ^= (color << 4);
						break;
					case 2:
						offScreenWindowBits[(((row + y) * screenWidth) + column + x) / 4] |= (0x3 << 2);
						offScreenWindowBits[(((row + y) * screenWidth) + column + x) / 4] ^= (color << 2);
						break;
					case 3:
						offScreenWindowBits[(((row + y) * screenWidth) + column + x) / 4] |= (0x3 << 0);
						offScreenWindowBits[(((row + y) * screenWidth) + column + x) / 4] ^= (color << 0);
						break;
				}
			}
		}
	}
}


/******************************************************************
 * OffscreenCopyAreaFromBuffer
 *
 * Copies a 8bit pixel array into the corresponding area of the
 * off-screen window. Probably can be made more efficient.
 *****************************************************************/
void OffscreenCopyAreaFromBuffer(CARD8 *buf, short x, short y, 
		unsigned short width, unsigned short height) 
{
	short row = 0;
	short column = 0;
	short start_column = 0;
	short end_column = 0;
	short start_row = 0;
	short end_row = 0;

	short x_off = 0;
	unsigned short color_temp = 0;
	unsigned short color_mask = 0;

	// align to real coordinates
	x -= palmFB.virtual.x;
	y -= palmFB.virtual.y;

	// Check top-left corner
	if ((x >= palmFB.virtual.w) || (y >= palmFB.virtual.h))	{
		return;
	}

	// Check top and bottom boundaries
	if (y < 0) {
		start_row = -y;
	}
	else {
		start_row = 0;
	}
	if ((y + height) >= palmFB.virtual.h) {
		end_row = (palmFB.virtual.h - y) - 1;
	}
	else {
		end_row = height-1;
	}
	if (end_row < start_row) {
		return;
	}

	// Check left and right boundaries
	if (x < 0) {
		start_column = -x;
	}
	else {
		start_column = 0;
	}
	if ((x + width) >= palmFB.virtual.w) {
		end_column = (palmFB.virtual.w - x) - 1;
	}
	else {
		end_column = width-1;
	}
	if (end_column < start_column) {
		return;
	}


	if (VNC_screenMode == modeSonyHires 
			|| VNC_screenMode == modeHiDensity
			|| VNC_screenMode == modeHandEra330) 
	{
	  WinHandle prevDrawWin = WinSetDrawWindow(offScreenWindow);
  	
		if (VNC_screenMode == modeHiDensity) {
			WinPushDrawState();
	  	WinSetCoordinateSystem(kCoordinatesNative);
	  }
	  	
		for (row = start_row; row <= end_row; row++) {
			for (column = start_column; column <= end_column; column ++) {
				VncSetForeColor(buf[(row * width) + column]);
				CommonWinDrawPixel(column + x, row + y);
			}
		}

		if (VNC_screenMode == modeHiDensity) {
			WinPopDrawState();
		}
		
  	WinSetDrawWindow(prevDrawWin);
	}
	else if (screenDepth == 8) {
		/* Do middle region */
		for (row = start_row; row <= end_row; row++) {
			for (column = start_column; column <= end_column; column ++) {
				offScreenWindowBits[(((row + y) * screenWidth) + column + x)] = buf[(row * width) + column];
			}
		}
	}
	else if (screenDepth == 4) {
		/* Do left-hand column if non-aligned */
		if (((x+column) % 2)) {  /* 1, 3, 5, ...  */
			for (row = start_row; row <= end_row; row++) {
				offScreenWindowBits[(((row + y) * screenWidth) + start_column + x) / 2] |= (0xF);
				offScreenWindowBits[(((row + y) * screenWidth) + start_column + x) / 2] ^= (BGR233ToIntensityNibble[buf[(row * width) + start_column]]);
			}
			start_column++;
		}

		/* Check if only 1 column */
		if (end_column < start_column) {
			return;
		}

		/* Do right-hand column if non-aligned */
		if (!((x+end_column) % 2)) {  /* 0, 2, 4, ...  */
			for (row = start_row; row <= end_row; row++) {
				offScreenWindowBits[(((row + y) * screenWidth) + end_column + x) / 2] |= (0xF << 4);
				offScreenWindowBits[(((row + y) * screenWidth) + end_column + x) / 2] ^= (BGR233ToIntensityNibble[buf[(row * width) + end_column]] << 4);
			}
			end_column--;
		}

		/* Check if only 2 columns */
		if (end_column <= start_column) {
			return;
		}

		/* Do middle region */
		for (row = start_row; row <= end_row; row++) {
			for (column = start_column; column < end_column; column += 2)	{
				offScreenWindowBits[(((row + y) * screenWidth) + column + x) / 2] = ~(BGR233ToIntensityNibble[buf[(row * width) + column + 1]] +
																			   (BGR233ToIntensityNibble[buf[(row * width) + column]] << 4));
			}
		}
	}
	else if (screenDepth == 2) {
		/* Very slow implementation for now to sanity-check */

		for (row = start_row; row <= end_row; row++) {
			for (column = start_column; column <= end_column; column++)	{
				switch ((x + column) % 4)	{
					case 0:
						offScreenWindowBits[(((row + y) * screenWidth) + column + x) / 4] |= (0x3 << 6);
						offScreenWindowBits[(((row + y) * screenWidth) + column + x) / 4] ^= ((BGR233ToIntensityNibble[buf[(row * width) + column]] << 4) & (0x3 << 6));
						break;
					case 1:
						offScreenWindowBits[(((row + y) * screenWidth) + column + x) / 4] |= (0x3 << 4);
						offScreenWindowBits[(((row + y) * screenWidth) + column + x) / 4] ^= ((BGR233ToIntensityNibble[buf[(row * width) + column]] << 2) & (0x3 << 4));
						break;
					case 2:
						offScreenWindowBits[(((row + y) * screenWidth) + column + x) / 4] |= (0x3 << 2);
						offScreenWindowBits[(((row + y) * screenWidth) + column + x) / 4] ^= ((BGR233ToIntensityNibble[buf[(row * width) + column]] << 0) & (0x3 << 2));
						break;
					case 3:
						offScreenWindowBits[(((row + y) * screenWidth) + column + x) / 4] |= (0x3 << 0);
						offScreenWindowBits[(((row + y) * screenWidth) + column + x) / 4] ^= ((BGR233ToIntensityNibble[buf[(row * width) + column]] >> 2) & (0x3 << 0));
						break;
				}
			}
		}
	}
}

/******************************************************************
 * CreateWindows
 *
 * Creates the off-screen panning window and inits global variables.
 * Called when the MainFrom is loaded and opened, and also
 * upon screen change notify events, if the screen size has
 * changed (silk screen display/hide)
 *****************************************************************/
static Err CreateWindows(FormPtr frmPtr)
{
  Err error;
  WinPtr offScreenWindowWinPtr;
  WinHandle prevDrawWin = WinGetDrawWindow();
  BitmapType* keyboard_bmp;
  MemHandle keyboard_res;


//	viewableMaxW = screenWidth;
//	viewableMaxH = screenHeight;
int	screenUsableWidth = screenWidth - VNC_prefs.scrollbarWidth;
int	screenUsableHeight = screenHeight - VNC_prefs.scrollbarWidth;
	

	// init viewport
  palmFB.viewable.x = palmFB.viewable.y = 0;
  palmFB.viewable.h = screenUsableHeight;
  palmFB.viewable.w = screenUsableWidth;

  palmFB.zoomarea_ena = false;

  palmFB.virtual.x = palmFB.virtual.y = 0;
	palmFB.virtual.h = screenUsableHeight;
	palmFB.virtual.w = screenUsableWidth;

//  palmFB.remote.x = palmFB.remote.y = palmFB.remote.h = palmFB.remote.w = 0;

  pen_zooming = false;

//	virtualBounds.topLeft.x = 0;
//	virtualBounds.topLeft.y = 0;
//	virtualBounds.extent.x = screenUsableWidth;
//	virtualBounds.extent.y = screenUsableHeight;

  viewportX = 0;
  viewportY = 0;

  /* set up the size of the update rectangle */
  updateRect.extent.x = palmFB.viewable.w;
  updateRect.extent.y = palmFB.viewable.h;


// To be allowed to draw in the silk screen area on Hires+ devices,
// we must use the display window and not the main form window, because
// form windows are limited by clipping to the application drawing area (320x320)

//  topLevel = FrmGetWindowHandle (frmPtr);
  topLevel = WinGetDisplayWindow();

  if (! offScreenWindow) {
		/* offScreenWindow = WinGetDrawWindow(); */
  	if (VNC_screenMode == modeHiDensity) {
  		// with the high-density API, we must use "standard" coordinates, which
  		// means that even though we pass e.g. 160x160, the resulting bitmap
  		// will be 320x320... 
	   	offScreenWindow = CommonWinCreateOffscreenWindow(screenWidth / density, screenHeight / density,
				nativeFormat, &error);
  	}
  	else {
	  	offScreenWindow = CommonWinCreateOffscreenWindow(screenWidth, screenHeight,
				screenFormat, &error);
		}
	}

  if (! keyboard) {
		keyboard_res = DmGetResource(bitmapRsc, KeyboardBitmapFamily);
		keyboard_bmp = MemHandleLock(keyboard_res);
		BmpGlueGetDimensions(keyboard_bmp, &keyboardWidth, &keyboardHeight, NULL);
		keyboard = WinCreateOffscreenWindow(keyboardWidth, keyboardHeight, screenFormat, &error);
		WinSetDrawWindow(keyboard);
		WinDrawBitmap(keyboard_bmp, 0, 0);
		MemHandleUnlock(keyboard_res);
		DmReleaseResource(keyboard_res);
  }

//  WinSetDrawWindow(offScreenWindow);
//  WinEraseWindow();
//  WinSetDrawWindow(prevDrawWin);

	if (VNC_screenMode == modeLegacy8bpp || VNC_screenMode == modeLegacy4bpp || VNC_screenMode == modeLegacy2bpp) {
	  offScreenWindowWinPtr = WinGetWindowPointer(offScreenWindow);
	 	//offScreenWindowBits = offScreenWindowWinPtr->displayAddrV20;
		offScreenWindowBits = BmpGetBits(WinGetBitmap(offScreenWindowWinPtr));
	}

	// init scrollbars  
	vertScrollbarRect.topLeft.x = screenWidth - VNC_prefs.scrollbarWidth;
	vertScrollbarRect.topLeft.y = 0;
	vertScrollbarRect.extent.x = VNC_prefs.scrollbarWidth;
	vertScrollbarRect.extent.y = screenHeight;

	horizScrollbarRect.topLeft.x = 0;
	horizScrollbarRect.topLeft.y = screenHeight - VNC_prefs.scrollbarWidth;
	horizScrollbarRect.extent.x = screenHeight - VNC_prefs.scrollbarWidth;
	horizScrollbarRect.extent.y = VNC_prefs.scrollbarWidth;

	vertSlideRect.topLeft.x = screenWidth - VNC_prefs.scrollbarWidth;
	vertSlideRect.topLeft.y = 0;
	vertSlideRect.extent.x = VNC_prefs.scrollbarWidth;
	vertSlideRect.extent.y = screenHeight;

	horizSlideRect.topLeft.x = 0;
	horizSlideRect.topLeft.y = screenHeight - VNC_prefs.scrollbarWidth;
	horizSlideRect.extent.x = screenWidth;
	horizSlideRect.extent.y = VNC_prefs.scrollbarWidth;

  return 0;
}


/******************************************************************
 * OpenNetworkPreferences
 *
 * Opens the network preferences dialog
 *****************************************************************/
static Boolean OpenNetworkPreferences(void) {
	Err                     err;
	UInt                    cardNo;
	LocalID                 dbID;
	DmSearchStateType       searchState;

	// Shut down the app, reset video, etc.
//  StopApplication();

	// Get the card number and database id of the Address application.
	err = DmGetNextDatabaseByTypeCreator(true, &searchState,
		sysFileTPanel, sysFileCNetworkPanel, true, &cardNo, &dbID);
	ErrFatalDisplayIf(err, "Network panel not found");
	if (err) {
		return false;
	}

	SysUIAppSwitch(cardNo, dbID, sysAppLaunchCmdPanelCalledFromApp, NULL);
	return true;
}


/******************************************************************
 * UpdateScrollbars
 *
 * Repaints scrollbars
 *****************************************************************/
static void UpdateScrollbars(void) {
  WinHandle prevDrawWin;

	if (VNC_screenMode == modeHiDensity) {
		WinPushDrawState();
		prevDrawWin = WinSetDrawWindow(WinGetDisplayWindow());
		WinSetCoordinateSystem(kCoordinatesNative);
	}

  WinSetForeColor(VNC_prefs.scrollBgColor);
  CommonWinDrawRectangle(&horizScrollbarRect, 0);
  CommonWinDrawRectangle(&vertScrollbarRect, 0);

  WinSetForeColor(VNC_prefs.scrollSliderColor);
  CommonWinDrawRectangle(&horizSlideRect, 0);
  CommonWinDrawRectangle(&vertSlideRect, 0);
	
  if (VNC_screenMode == modeHiDensity) {
  	WinSetDrawWindow(prevDrawWin);
  	WinPopDrawState();
  }
	
}


/******************************************************************
 * PositionScrollbars
 *
 * re-computes scrollbar rectangles and re-paints scrollbars
 *****************************************************************/
static void PositionScrollbars()
{
	WinSetDrawWindow(topLevel);

  if (palmFB.remote.w == 0) {
		horizSlideRect.topLeft.x = 0;
	} else {
		horizSlideRect.topLeft.x = (((UInt32) viewportX * (UInt32) palmFB.viewable.w) / palmFB.remote.w);
	}

  if (palmFB.remote.h == 0) {
		vertSlideRect.topLeft.y = 0;
	} else {
		vertSlideRect.topLeft.y = ((UInt32) (viewportY * (UInt32) palmFB.viewable.h) / palmFB.remote.h);
	}

  UpdateScrollbars();
}


/******************************************************************
 * ShutdownConnection
 *
 * Closes the current connection to the VNC currentServer
 *****************************************************************/
static void ShutdownConnection() {
	Err err;
	
  if (rfbsock > 0) {
		
		NetLibSocketClose(AppNetRefnum, rfbsock, ConnectNetTimeout, &err);
		rfbsock = 0;

		PrintSentRecvStat();

		printf("Connection closed\n\n");

		/* set up the size of scrollbar slides */
		horizSlideRect.extent.x = 0;
		vertSlideRect.extent.y = 0;

		viewportX = palmFB.viewable.x = palmFB.virtual.x = 0;
		viewportY = palmFB.viewable.y = palmFB.virtual.y = 0;

		palmFB.remote.h = palmFB.remote.w = 0;

		/*InitVirtual();*/

		PositionScrollbars();
	//  UpdateViewport();

		PostProgressMessage("Connection closed", 0, 0);
  }
}


/******************************************************************
 * ShutdownNetwork
 *
 * Closes the network. If enforcedClose == True requests the modem
 * to hang up, otherwise closes only ots own API to the Network lib.
 *****************************************************************/
static void ShutdownNetwork(Boolean enforcedClose) {
  Err err;
  Char AppErrStr[32];

  ShutdownConnection();

  printf("Closing NetLib...\n");
  err = NetLibClose(AppNetRefnum, enforcedClose);

  if (err == netErrNotOpen) {
		printf("ERROR: cannot close: %s\n", 
			SysErrString(err,AppErrStr,sizeof(AppErrStr)));
	}
  else if (err == netErrStillOpen) {
		printf("Network is still in use\n");
	}
  else {
		printf("Network is closed\n");
	}

  networkIsOpen = false;
  return;
}


/******************************************************************
 * OpenNetwork
 *
 * Loads the lib and connects to the network
 *****************************************************************/
static Boolean OpenNetwork(void) {
  UInt16  ifErrs;
  Err  err;
  Char AppErrStr[32];

  printf("\nOpening the network...");

  err = NetLibOpen(AppNetRefnum, &ifErrs);

	if (err == netErrAlreadyOpen || err == 0) {
		printf("skipped\n");
		return true;
	}	
	else {
		if (ifErrs == netErrUserCancel) {
			ShutdownNetwork(true);
			return false;
		}
		printf("\nERROR: opening network: %s\n", 
			SysErrString(err,AppErrStr,sizeof(AppErrStr)));

		if (ifErrs == netErrPrefNotFound) {
			printf("\nNetwork preference not set\n");
		}

		return false;
	}

	printf("opened\n");

	return true;
}


/******************************************************************
 * OpenConnection
 *
 * Opens a connection to the currentServer
 *****************************************************************/
static Boolean OpenConnection(void)
{
  Err  err;
  int port;

	if (rfbsock > 0) {
		printf("Already connected\n");
		return true;
	}

	if (! networkIsOpen) {
	  if ((networkIsOpen = OpenNetwork()) != true) {
		  printf("ERROR: cannot open network\n");
		  return false;
	  }
	}

	port = currentServer.basePort + currentServer.displayNumber;
	
  printf("Connecting to %s:%d...\n", currentServer.address, port);

  err = 0;

  // always use BGR233 format! Maybe we will have something else in the future
  useBGR233 = True;

  /* this is a set-up for BRG233 format. We tell the currentServer that we
   * use BRG233 and have 8 bit color depth. We map BRG233 pixels to
   * 4 bit grayscale. All that is because VNC currently does not
   * provide data formats for depths less then 8
   */
  myFormat.bitsPerPixel = 8;
  myFormat.depth = 8;
  myFormat.trueColour = 1;
  myFormat.bigEndian = 0;
  myFormat.redMax = 7;
  myFormat.greenMax = 7;
  myFormat.blueMax = 3;
  myFormat.redShift = 0;
  myFormat.greenShift = 3;
  myFormat.blueShift = 6;

  viewportX = palmFB.viewable.x = palmFB.virtual.x = currentServer.beginX;
  viewportY = palmFB.viewable.y = palmFB.virtual.y = currentServer.beginY;

  PostProgressMessage("Connecting...", 0, 0);

  if (!ConnectToRFBServer(currentServer.address, port)) {
		printf("ERROR: cannot connect to %s:%d\n", currentServer.address, port);
		err = true;
  } else if (!InitialiseRFBConnection(rfbsock)) {
		printf("ERROR: cannot initialize %s:%d\n", currentServer.address, port);
		err = true;
  } else if (!SetFormatAndEncodings()) {
		err = true;
		printf("ERROR: cannot set format\n");
  } else if (currentServer.scaleFactor != 1) {
		if (!SetScaleFactor(currentServer.scaleFactor)) {
			err = true;
		}
  }
  
  if (currentServer.scaleFactor == 1) {
	  if ((err == false) 
	  		&& (!SendFramebufferUpdateRequest(
	  			palmFB.virtual.x, palmFB.virtual.y,
					palmFB.virtual.w, palmFB.virtual.h, 
					False))) 
		{
			printf("ERROR: cannot send buffer update\n");
			err = true;
	  }
  }

  if (err) {
		PostProgressMessage("Cannot connect!", 0, 0);
		ShutdownConnection();
		return false;
  }

  /* set up the size of scrollbar slides */
  horizSlideRect.extent.x =
		((Int32) palmFB.viewable.w * horizScrollbarRect.extent.x) / palmFB.remote.w;
  vertSlideRect.extent.y =
		((Int32) palmFB.viewable.h * vertScrollbarRect.extent.y) / palmFB.remote.h;

  /* set the pen time tracing, used to catch double-taps */
  lastPenTime = TimGetTicks();

  printf("Connected!\n");

  return true;
}


/******************************************************************
 * SendKeySequence
 *
 * Sends a sequence of n characters as a keyboard input to the VNC
 * currentServer. Works with unsigned short data because the server works
 * with 16bit KeySyms from X11. mod1 and mod2 are two available
 * modifiers. If ascii is true, convert the data into a sequence of
 * unsigned short
 *****************************************************************/
static void SendKeySequence(unsigned short mod1, unsigned short mod2, unsigned short mod3,
			  unsigned short *keys, int n, Boolean ascii)
{
  int i;
  unsigned short key;

  if (keys == 0) {
		return;
  }

  if (mod1 > 0) {
		SendKeyEvent(mod1, true); // modifier 1 pressed
	}

  if (mod2 > 0) {
		SendKeyEvent(mod2, true); // modifier 2 pressed
	}

  if (mod3 > 0) {
		SendKeyEvent(mod3, true); // modifier 3 pressed
	}

  for (i = 0; i < n; i++) {
		if (ascii) {
		  key = ((unsigned char *)keys)[i];
		}	else {
		  key = keys[i];
		}

		// Overrides for some controls keys
	  if (key == chrLineFeed) {
	  	key = XK_Return; 
	  } else if (key == chrBackspace) {
	  	key = XK_BackSpace;
	  } else if (key == chrTab) {
	  	key = XK_Tab;
	  } else if (key == chrLeftArrow) {
	  	key = XK_Left;
	  } else if (key == chrRightArrow) {
	  	key = XK_Right;
	  } else if (key == chrUpArrow) {
	  	key = XK_Up;
	  } else if (key == chrDownArrow) {
	  	key = XK_Down;
	  } 
	  	
		// process shift
		if ((keyboard_state & (1 << KEYBOARD_VISIBLE)) && (keyboard_state & (1 << KEYBOARD_SHIFT)))	{
			if ((key >= 'a') && (key <= 'z')) {
				key -= ('a' - 'A');
			}
		}

		// process function key
		if ((keyboard_state & (1 << KEYBOARD_VISIBLE)) && (keyboard_state & (1 << KEYBOARD_FN))) {
			if ((key >= '1') && (key <= '9')) {
				key = (key - '1') + XK_F1;
			}
			if (key == '0') {
				key = XK_F1 + 9;
			}
		}

		SendKeyEvent(key, true); // pressed
		SendKeyEvent(key, false); // released
  }

  if (mod3 > 0) {
		SendKeyEvent(mod3, false); // released
	}

  if (mod2 > 0) {
		SendKeyEvent(mod2, false); // released
	}

  if (mod1 > 0) {
		SendKeyEvent(mod1, false); // released
	}
}


/******************************************************************
 * UpdateViewport
 *
 * Copies the viewable portion of the off-screen window into the
 * visible window
 *****************************************************************/
static void UpdateViewport(void)
{
  RectangleType rect;
  int i;

	if (DrawingBlocked()) {
  	return;
  }

  if (rfbsock == 0) {
  	//InitScreen();
		return;
	}

  if (VNC_screenMode == modeHiDensity) {
  	WinPushDrawState();
  	WinSetCoordinateSystem(kCoordinatesNative);
  }
  CommonWinCopyRectangle(offScreenWindow, topLevel,
		GetUpdateRectAtPos(palmFB.viewable.x - palmFB.virtual.x,
												palmFB.viewable.y - palmFB.virtual.y),
		0, 0, scrCopy);
  if (VNC_screenMode == modeHiDensity) {
  	WinPopDrawState();
  }

	if (keyboard_state & (1 << KEYBOARD_VISIBLE))	{
		keyboardY = (screenHeight - VNC_prefs.scrollbarWidth) / density 
									- keyboardHeight / 2;
		/* Copy keyboard */
		rect.extent.x = 1 + KEYBOARD_KEY_WIDTH;
		rect.extent.y = 6;
		for (i = 0; i < KEYBOARD_KEY_COUNT; i++) {
			rect.topLeft.x = KEYBOARD_KEY_WIDTH * i;
			if (keyboard_state & (1 << i)) {
				rect.topLeft.y = 6;
			}
			else {
				rect.topLeft.y = 0;
			}
			WinCopyRectangle(keyboard, topLevel, &rect, rect.topLeft.x, keyboardY, scrCopy);
		}
	}
}


/******************************************************************
 * PositionViewport
 *
 * Moves visible area over the virtual panning window. If the
 * the boundaries of the visible area is beyond the panning window,
 * the panning window will be moved and a new portion of the screen
 * data will be requested
 *****************************************************************/
static void PositionViewport()
{
  int deltaX=0, deltaY=0;
  int updateViewport = False;

  if (viewportX < 0) {
		viewportX = 0;
	}

  if (viewportX > (palmFB.remote.w - palmFB.viewable.w)) {
		viewportX = palmFB.remote.w - palmFB.viewable.w;
	}

  if (viewportY < 0) {
		viewportY = 0;
	}

  if (viewportY > (palmFB.remote.h - palmFB.viewable.h)) {
		viewportY = palmFB.remote.h - palmFB.viewable.h;
	}

  if (palmFB.viewable.x != viewportX) {
		if (viewportX < palmFB.virtual.x ||
				viewportX + palmFB.viewable.w > palmFB.virtual.w) {
		  deltaX = (int) palmFB.viewable.x - viewportX;
		}
		updateViewport = True;
		palmFB.viewable.x = viewportX;
  }

  if ((int) palmFB.virtual.x - deltaX < 0) {
		deltaX = palmFB.virtual.x;
	}

  if (palmFB.viewable.y != viewportY) {
		if (viewportY < palmFB.virtual.y ||
				viewportY + palmFB.viewable.h > palmFB.virtual.h) {
		  deltaY = (int)palmFB.viewable.y - viewportY;
		}
		updateViewport = True;
		palmFB.viewable.y = viewportY;
  }

  if ((int) palmFB.virtual.y - deltaY < 0) {
		deltaY = palmFB.virtual.y;
	}

  if (deltaX != 0 || deltaY != 0) {
		int obscuredRectX = 0, obscuredRectY = 0;
		RectangleType r;

		r.topLeft.x = r.topLeft.y = 0;
		r.extent.x = palmFB.virtual.w;
		r.extent.y = palmFB.virtual.h;

		if (deltaX > 0) {
			r.extent.x -= deltaX;
		}
		if (deltaY > 0) {
			r.extent.y -= deltaY;
		}

		
		if (VNC_screenMode == modeHiDensity) {
			WinPushDrawState();
			WinSetCoordinateSystem(kCoordinatesNative);
		}
		WinCopyRectangle(offScreenWindow, offScreenWindow, &r, deltaX, deltaY, scrCopy);
		if (VNC_screenMode == modeHiDensity) {
			WinPopDrawState();
		}

		if (deltaX < 0) {
		  obscuredRectX = (int) palmFB.virtual.w + deltaX;
		}

		if (deltaX != 0) {
			int absDeltaX = deltaX;
			if (absDeltaX < 0) {
				absDeltaX = - absDeltaX;
			}

		  OffscreenDrawFilledRectangle(LIGHT_GRAY, obscuredRectX, 0,
				absDeltaX, palmFB.virtual.h, false);

		  palmFB.virtual.x -= deltaX;
		  deltaX = absDeltaX;
		  if (deltaX > 156 * density) {
		  	deltaX = 156 * density;
		  }

		  SendFramebufferUpdateRequest(
		  	palmFB.virtual.x + obscuredRectX,	palmFB.virtual.y,
				deltaX,	palmFB.virtual.h,
				False);
		}

		if (deltaY < 0) {
		  obscuredRectY = (int)palmFB.virtual.h + deltaY;
		}

		if (deltaY != 0) {
			int absDeltaY = deltaY;
			if (absDeltaY < 0) {
				absDeltaY = - absDeltaY;
			}

		  OffscreenDrawFilledRectangle(LIGHT_GRAY, 0, obscuredRectY,
				palmFB.virtual.w, absDeltaY, false);

		  palmFB.virtual.y -= deltaY;
		  deltaY = absDeltaY;
		  if (deltaY > 156 * density) {
		  	deltaY = 156 * density;
		  }

		  SendFramebufferUpdateRequest(
		  	palmFB.virtual.x,	palmFB.virtual.y + obscuredRectY,
				palmFB.virtual.w,	deltaY,
				False);
		}

		sendUpdateRequest = true;
		updateViewport = true;
  }

  if (updateViewport) {
		UpdateViewport();
	}
}


/******************************************************************
 * PointInRectangle
 *
 * returns true if the point (x,y) is in the rect. rp otherwise
 * false
 *****************************************************************/
static Boolean PointInRectangle(unsigned x, unsigned y, RectangleType *rp) {
  return ((x >= rp->topLeft.x) &&
		(y >= rp->topLeft.y) &&
		((unsigned) (x - rp->topLeft.x) < rp->extent.x) &&
		((unsigned) (y - rp->topLeft.y) < rp->extent.y));
}


/******************************************************************
 * InMainArea
 *
 * returns true if the point (x,y) is in the VNC area otherwise
 * false
 *****************************************************************/
static Boolean InMainArea(unsigned short x, unsigned short y) {
  return (x < palmFB.viewable.w && y < palmFB.viewable.h);
}

/******************************************************************
 * InKeyboardArea
 *
 * returns true if the point (x,y) is in the keyboard area otherwise
 * false
 *****************************************************************/
static Boolean InKeyboardArea(unsigned short x, unsigned short y) {
	if (!(keyboard_state & (1 << KEYBOARD_VISIBLE))) {
		return false;
	}
		
	if ((x < (KEYBOARD_KEY_COUNT * KEYBOARD_KEY_WIDTH + 1) * 2) 
			&& (y >= keyboardY*density) && (y <= (keyboardY+keyboardHeight)*density)) 
	{
		return true;
	}
		
	return false;
}


/******************************************************************
 * ToggleKeyboardState
 *
 * Toggles the special keyboard on and off
 ******************************************************************/
static void ToggleKeyboardState(void)
{
	keyboard_state ^= (1 << KEYBOARD_VISIBLE);
	if (keyboard_state & (1 << KEYBOARD_VISIBLE)) {
		if (keyboard_state & (1 << KEYBOARD_SHIFT)) {
			SendKeyEvent(XK_Shift_L, true);
		}
		if (KEYBOARD_CTRL & (1 << KEYBOARD_CTRL)) {
			SendKeyEvent(XK_Control_L, true);
		}
		if (KEYBOARD_ALT & (1 << KEYBOARD_ALT)) {
			SendKeyEvent(XK_Alt_L, true);
		}
	}
	else {
		if (keyboard_state & (1 << KEYBOARD_SHIFT)) {
			SendKeyEvent(XK_Shift_L, false);
		}
		if (keyboard_state & (1 << KEYBOARD_CTRL)) {
			SendKeyEvent(XK_Control_L, false);
		}
		if (keyboard_state & (1 << KEYBOARD_ALT)) {
			SendKeyEvent(XK_Alt_L, false);
		}
	}
}



/******************************************************************
 * TranslateEventCoords
 *
 * Scale Event coordinates so that they match the actual screen coordinates
 ******************************************************************/
static void TranslateEventCoords(EventPtr e) {
	e->screenX *= density;
	e->screenY *= density;
}



/******************************************************************
 * ProcessKeyboardPenDownEvent
 *
 * Handles events penDown for the keyboard masked area
 ******************************************************************/
static void ProcessKeyboardPenDownEvent(EventPtr e)
{
	unsigned short key;

//	TranslateEventCoords(e);

	if (e->screenX >= 156 * density) {
		ToggleKeyboardState();
	}
	else {
		key = (e->screenX / (KEYBOARD_KEY_WIDTH * density));
		keyboard_state ^= (1 << key);
		if (key == KEYBOARD_SHIFT) {
			if (keyboard_state & (1 << KEYBOARD_SHIFT)) {
				SendKeyEvent(XK_Shift_L, true);
			}	else {
				SendKeyEvent(XK_Shift_L, false);
			}
		}
		if (key == KEYBOARD_CTRL)	{
			if (keyboard_state & (1 << KEYBOARD_CTRL)) {
				SendKeyEvent(XK_Control_L, true);
			}	else {
				SendKeyEvent(XK_Control_L, false);
			}
		}
		if (key == KEYBOARD_ALT) {
			if (keyboard_state & (1 << KEYBOARD_ALT)) {
				SendKeyEvent(XK_Alt_L, true);
			}	else {
				SendKeyEvent(XK_Alt_L, false);
			}
		}
	}

	UpdateViewport();
}

/******************************************************************
 * ProcessKeyboardPenUpEvent
 *
 * Handles events penUp for the keyboard masked area
 ******************************************************************/
static void ProcessKeyboardPenUpEvent(EventPtr e)
{
	unsigned short key;
	unsigned short oneKey;

	if (e->screenX >= 156 * density) {
		return;
	}

//	TranslateEventCoords(e);

	key = (e->screenX / (KEYBOARD_KEY_WIDTH * density));

	/* Escape key */
	if (key == 0)	{
		keyboard_state ^= (1 << key);
		oneKey = XK_Escape;
		SendKeySequence(0, 0, 0, &oneKey, 1, false);
		UpdateViewport();
	}
}


/******************************************************************
 * InScrollbar
 *
 * returns true if the point (x,y) is in the specified scrollbar
 * area otherwise false
 *****************************************************************/
static Boolean InScrollbar(unsigned short barType, unsigned short x, unsigned short y) {
  Boolean ret = false;

  switch (barType) {
	case HORIZ:
	  ret = PointInRectangle(x, y, &horizSlideRect);
	  break;

	case HORIZ_UP:
	  ret = PointInRectangle(x, y, &horizScrollbarRect) &&
			! PointInRectangle(x, y, &horizSlideRect) &&
			(x < horizSlideRect.topLeft.x);
	  break;

	case HORIZ_DOWN:
	  ret = PointInRectangle(x, y, &horizScrollbarRect) &&
			! PointInRectangle(x, y, &horizSlideRect) &&
			(x > horizSlideRect.topLeft.x + horizSlideRect.extent.x);
	  break;

	case HORIZ_TOP:
	  ret = false;
	  break;

	case HORIZ_BUTTOM:
	  ret = false;
	  break;

	case VERT:
	  ret = PointInRectangle(x, y, &vertSlideRect);
	  break;

	case VERT_UP:
	  ret = PointInRectangle(x, y, &vertScrollbarRect) &&
			!PointInRectangle(x, y, &vertSlideRect) &&
			(y < vertSlideRect.topLeft.y);
	  break;

	case VERT_DOWN:
	  ret = PointInRectangle(x, y, &vertScrollbarRect) &&
			!PointInRectangle(x, y, &vertSlideRect) &&
			(y > vertSlideRect.topLeft.y + horizSlideRect.extent.y);
	  break;

	case VERT_TOP:
	  ret = false;
	  break;

	case VERT_BUTTOM:
	  ret = false;
	  break;

	default:
	  return false;
  }

  return ret;
}


/******************************************************************
 * ProcessScrollPenMoveEvent
 *
 * Handles events penMoveEvent in the main form
 *****************************************************************/
static void ProcessScrollPenMoveEvent(EventPtr e)
{
//	TranslateEventCoords(e);

  if (buttonState == HORIZ) {
		viewportX += ((e->screenX - downX) 
			* (int) (palmFB.remote.w / palmFB.viewable.w));
		downX = e->screenX;
  } 
  else if (buttonState == VERT) {
		viewportY += ((e->screenY - downY) 
			* (int) (palmFB.remote.h	/ palmFB.viewable.h));
		downY = e->screenY;
  } 
  else {
		return;
	}

  PositionViewport();
  PositionScrollbars();
}


/******************************************************************
 * ProcessScrollKeyEvent
 *
 * Handles events keyEvent for the scrolling hard keys
 * Returns true is the event was processed as a scroll event
 ******************************************************************/
static Bool ProcessScrollKeyEvent(EventPtr e)
{
	WChar chr = e->data.keyDown.chr;
	int pixelShift = 0x10 << VNC_prefs.screenPanningProp;
	
	if (chr == vchrPageUp
			|| chr == vchrJogUp)
	{
		viewportY -= pixelShift;
		return true;
	}
	if (chr == vchrPageDown
			|| chr == vchrJogDown) 
	{
		viewportY += pixelShift;
		return true;
	}
	if (chr == vchrHard3
			|| chr == vchrRockerLeft
			|| NavKeyPressed(e, Left)) 
	{
		viewportX -= pixelShift;
		return true;
	}
	if (chr == vchrHard4
			|| chr == vchrRockerRight
			|| NavKeyPressed(e, Right))
	{
		viewportX += pixelShift;
		return true;
	}
	
	return false;
}


/******************************************************************
 * ProcessScrollPenDownEvent
 *
 * Handles events penDownEvent in scrollers
 *****************************************************************/
static void ProcessScrollPenDownEvent(EventPtr e)
{
//	TranslateEventCoords(e);

  if (InScrollbar(HORIZ, e->screenX , e->screenY)) {
		downX =  e->screenX;
		buttonState = HORIZ;
		return;
  } 
  else if (InScrollbar(VERT,  e->screenX ,  e->screenY)) {
		downY =  e->screenY;
		buttonState = VERT;
		return;
  }
  else if (InScrollbar(HORIZ_UP,  e->screenX ,  e->screenY)) {
		viewportX -= palmFB.viewable.w;

  } 
  else if (InScrollbar(HORIZ_DOWN, e->screenX , e->screenY)) {
		viewportX += palmFB.viewable.w;

  }
  else if (InScrollbar(VERT_UP, e->screenX , e->screenY)) {
		viewportY -= palmFB.viewable.h;

  } 
  else if (InScrollbar(VERT_DOWN, e->screenX , e->screenY)) {
		viewportY += palmFB.viewable.h;

  } 
  else {
		return;
	}

  buttonState = 0;
  PositionViewport();
  PositionScrollbars();
}


/******************************************************************
 * ProcessPenMoveEvent
 *
 * Handles events penMoveEvent in the main form
 *****************************************************************/
static void ProcessPenMoveEvent(EventPtr e)
{
  unsigned long penTime = TimGetTicks();
  unsigned short serverX, serverY;

//	TranslateEventCoords(e);

  if (DrawingBlocked()) {
		return;
	}

  if (penTime <= lastPenTime + 100) {
		return;
	}

  serverX = e->screenX + palmFB.viewable.x;
  serverY = e->screenY + palmFB.viewable.y;

  if (buttonState == MAIN_PRESSED || buttonState == MAIN_MOVED) {
		SendPointerEvent(serverX, serverY, currentMouseButton);
		buttonState = MAIN_MOVED;
  }
}


/******************************************************************
 * ProcessPenUpDownEvent
 *
 * Handles events penUpEvent and PenDownEvent in the main form
 *****************************************************************/
static void ProcessPenUpDownEvent(EventPtr e)
{
	unsigned long penTime;
	unsigned short serverX, serverY;

	if (DrawingBlocked()) {
		return;
	}
		
//	TranslateEventCoords(e);
		
	penTime = TimGetTicks();
	serverX = e->screenX + palmFB.viewable.x;
	serverY = e->screenY + palmFB.viewable.y;

	if (lastX == 0) {
		lastX = serverX;
	}
	if (lastY == 0) {
		lastY = serverY;
	}

	if (e->penDown) {
		if (pen_zooming == true) {
			palmFB.zoomarea_ena = true;
			palmFB.zoomarea.x = e->screenX;
			palmFB.zoomarea.y = e->screenY;
			/* viewportY = palmFB.viewable.y = palmFB.virtual.y = e->screenY; */
			SetScaleFactor(currentServer.scaleFactor);
			pen_zooming = false;
			return;
		}
		else {
			if (buttonState == MAIN_ONCE && penTime <= lastPenTime + 100)	{
				buttonState = MAIN_DOUBLE;
				SendPointerEvent(lastX, lastY, currentMouseButton);
				SendPointerEvent(lastX, lastY, 0);
			}
			else {
				buttonState = MAIN_PRESSED;
				// sync the intial mouse position first
				SendPointerEvent(serverX, serverY, 0);
				SendPointerEvent(serverX, serverY, currentMouseButton);
			}

			lastX = serverX;
			lastY = serverY;
			return;
		}
	}

	else {   // penUp
		if (buttonState == MAIN_PRESSED) {
			SendPointerEvent(lastX, lastY, 0);
			buttonState = MAIN_ONCE;
		}
		else if (buttonState == MAIN_MOVED) {
			buttonState = 0;
			SendPointerEvent(serverX, serverY, 0);
		}


		// Allow the masked off key to now send keyDownEvents.
		KeySetMask(keyBitsAll);
		currentMouseButton = 1;
	}

	lastPenTime = penTime;
}




/******************************************************************
 * ConsoleFormHandleEvent
 *
 * Handles events in the Console form
 *****************************************************************/
static Boolean ConsoleFormHandleEvent(EventPtr e) {
	Boolean handled;
	FormPtr frm;
	UInt16 field;

	handled = false;
	frm = FrmGetActiveForm();

	// Let StdIO handler to it's things first.
	if (StdHandleEvent(e)) {
	  return true;
	}

	switch (e->eType) {
	case frmOpenEvent:
		SizeForm(frm);
		CollapseSetState(frm, collapseStateUser);
		FrmDrawForm(frm);
		handled = true;
		break;

	case ctlSelectEvent:
		switch (e->data.ctlSelect.controlID) {
		  case ConsoleCloseButton:
				StdSave();
				FrmReturnToForm(MainForm);
				handled = true;
			break;
		  case ConsoleClearButton:
				StdClear();
				handled = true;
				break;
		  default:
				handled = true;
				break;
		}
		break;

	case menuEvent:
		MenuEraseStatus(NULL);
		field = FrmGetFocus(frm);
		switch (e->data.menu.itemID) {
		  case EditCopy:
				FldCopy(FrmGetObjectPtr(frm, field));
				break;

		  case EditPaste:
				FldPaste(FrmGetObjectPtr(frm, field));
				break;

		  default:
				break;
		}
		break;

	default:
		break;
	}

	return handled;
}


/******************************************************************
 * MainFormHandleEvent
 *
 *
 *****************************************************************/
static Boolean MainFormHandleEvent(EventPtr e)
{
	Boolean handled;
	FormPtr frm;
	unsigned short oneKey;
	UInt16 clipboardLength;
	VoidHand handle;
	VoidPtr ptr;
	int scale;

	char mypass[32];

	handled = false;

	switch (e->eType) {
	
	case frmOpenEvent:
		frm = FrmGetActiveForm();
		
		FrmCopyLabel(frm, VersionLabel, palmvncVersion);
		FrmCopyLabel(frm, BuildLabel, BUILD);
		CollapseSetState(frm, collapseStateUser);
		SizeForm(frm);
		if (VNC_screenMode == modeHandEra330) {
			VgaFormModify(frm, vgaFormModify160To240);
		}
		WinEraseWindow();
		FrmDrawForm(frm);
		CreateWindows(frm);
		//UpdateScrollbars();
		//UpdateViewport();
		handled = true;
		break;

	case frmUpdateEvent:
		if (rfbsock > 0) {
			UpdateViewport();
			UpdateScrollbars();
		} else {
			frm = FrmGetActiveForm();
			FrmDrawForm(frm);
		}
		handled = true;
		break;

	case nilEvent:
		UpdateViewport();
		break;

	case ctlSelectEvent:
		switch (e->data.ctlSelect.controlID) {
		  case MainConnectButton:
				if (DoServerSelect())	{
					if (rfbsock > 0) {
						ShutdownConnection();
					}
					if (! OpenConnection()) {
			  		FrmPopupForm(ConsoleForm);
			  	}
					else {
						PositionScrollbars();
					}
				}
				handled = true;
				break;
		}
		break;

	case menuEvent:
		/*MenuEraseStatus(NULL);*/
		switch(e->data.menu.itemID) {
		  case ViewOpenConsole:
				FrmPopupForm(ConsoleForm);
				UpdateViewport();
				break;

		  case ServerConnect:
				if (DoServerSelect())	{
					if (rfbsock > 0) {
						ShutdownConnection();
					}
					if (! OpenConnection()) {
			  		FrmPopupForm(ConsoleForm);
			  	}
					else {
						PositionScrollbars();
					}
				}
				break;

		  case ServerDisconnect:
				ShutdownConnection();
				break;

		  case ServerHangup:
				ShutdownNetwork(true);
				PostProgressMessage("Modem disconnected", 0, 0);
				break;

		  case ServerSystemNetworkprefs:
				OpenNetworkPreferences();
				break;
				
			case ServerPalmVNCprefs:
				DoPrefsForm();
				UpdateViewport();
				break;
				
			case ServerDefineMacros:
				FrmAlert(NotImplementedYetAlert);
				//DoMacroList();
				//UpdateViewport();
				break;
				
			case ServerDefineKeyMappings:
				DoKeyMapList();
				UpdateViewport();
				break;
				
		  case HelpAbout:
				FrmCustomAlert(AboutAlert, palmvncVersion, NULL, NULL);
				break;

		  case HelpUsingPalmVNC:
				FrmHelp(HelpGeneralString);
				break;

		  case HelpRegistration:
				FrmHelp(HelpRegistrationString);
				break;

		  case HelpLicence:
				FrmHelp(HelpLicenceString);
				break;

		  // request a "hard" update of the whole virtual area
		  case ViewRefresh:
				SendFramebufferUpdateRequest(
					palmFB.virtual.x, palmFB.virtual.y,
					palmFB.virtual.w,	palmFB.virtual.h,
					false);
				break;

		  case ViewSavePosition:
				currentServer.beginX = palmFB.viewable.x;
				currentServer.beginY = palmFB.viewable.y;
				break;

		  case SendCtrlAltDel:
				oneKey = XK_Delete;
				SendKeySequence(XK_Control_L, XK_Alt_L, 0, &oneKey, 1, false);
				break;

		  case ViewSpecialKeys:
				ToggleKeyboardState();
				break;

		  case SendAltF4:
				oneKey = XK_F4;
				SendKeySequence(XK_Alt_L, 0, 0, &oneKey, 1, false);
				break;

		  case SendNTlogin:
				// switch to the user name text field
				oneKey = (unsigned char)'u';
				SendKeySequence(XK_Alt_L, 0, 0, &oneKey, 1, false);

				// send the user name
				if (StrLen(currentServer.loginNameNT) > 0) {
				  SendKeySequence(0, 0, 0, (unsigned short *) &currentServer.loginNameNT,
						StrLen(currentServer.loginNameNT), true);

				  // switch to the password text field
				  oneKey = (unsigned char)'p';
				  SendKeySequence(XK_Alt_L, 0, 0, &oneKey, 1, false);

				  // send the password
					if (StrLen(currentServer.passwNT) == 0) {
						MemSet(mypass, 32, 0);
						GetPassword(mypass, 32);
					}
					else {
						StrCopy(mypass, currentServer.passwNT);
					}
				  if (StrLen(mypass) > 0) {
						SendKeySequence(0, 0, 0, (unsigned short *) &mypass, StrLen(mypass), true);
					}
				}

				// confirm with "enter", even if no user name or password
				oneKey = XK_Return;
				SendKeySequence(0, 0, 0, &oneKey, 1, false);

				break;

		  case SendNTlogout:
				oneKey = (unsigned char) 'l';
				SendKeySequence(XK_Alt_L, 0, 0, &oneKey, 1, false);
				oneKey = XK_Return;
				SendKeySequence(0, 0, 0, &oneKey, 1, false);
				break;

		  case SendNTlock:
				oneKey = (unsigned char) 'w';
				SendKeySequence(XK_Alt_L, 0, 0, &oneKey, 1, false);
				break;

		  case SendNTreboot:
				oneKey = (unsigned char) 's';
				SendKeySequence(XK_Alt_L, 0, 0, &oneKey, 1, false);
				oneKey = (unsigned char) 'r';
				SendKeySequence(XK_Alt_L, 0, 0, &oneKey, 1, false);
				oneKey = XK_Return;
				SendKeySequence(0, 0, 0, &oneKey, 1, false);
				break;

		  case SendWinKey:
				oneKey = XK_Escape;
				SendKeySequence(XK_Control_L, 0, 0, &oneKey, 1, false);
				break;
				
			case SendRightClick:
				currentMouseButton = 4;
				break;

		  case SendPalmClipboard:
				handle = ClipboardGetItem(clipboardText, &clipboardLength);
				if (handle != 0 && clipboardLength != 0) {
				  ptr = MemHandleLock(handle);
				  SendClientCutText(ptr, clipboardLength);
				  MemHandleUnlock(handle);
				}
				break;

		  case ViewWholeDesktop:
		  case ViewScale11:
		  case ViewScale12:
		  case ViewScale13:
		  case ViewScale14:
				palmFB.zoomarea_ena = true;
				palmFB.zoomarea.x = viewportX + (palmFB.viewable.w / 2);
				palmFB.zoomarea.y = viewportY + (palmFB.viewable.h / 2);

				switch (e->data.menu.itemID) {
					case ViewWholeDesktop:
						scale = (palmFB.desktop.w / palmFB.viewable.w);
						if (palmFB.desktop.w > (palmFB.viewable.w * scale)) {
							scale++;
						}
						if ((palmFB.desktop.h / palmFB.viewable.h) > scale) {
							scale = (palmFB.desktop.h / palmFB.viewable.h);
						}
						if (palmFB.desktop.h > (palmFB.viewable.h * scale)) {
							scale++;
						}
						pen_zooming = true;
						break;

					case ViewScale12:   
						scale = 2;  
						break;
						
					case ViewScale13:   
						scale = 3;  
						break;
						
					case ViewScale14:   
						scale = 4;  
						break;
						
					default:                            
						scale = 1;  
						break;
				}

				SetScaleFactor(scale);
				break;

		  default:
				break;
		}
		handled = true;
		break;

	case penDownEvent:
		if (rfbsock <= 0) {
			break;
		}
		
		TranslateEventCoords(e);

		if (InKeyboardArea(e->screenX, e->screenY)) {
			ProcessKeyboardPenDownEvent(e);
			handled = true;
		}
		else if (!InMainArea(e->screenX, e->screenY)) {
			ProcessScrollPenDownEvent(e);
			handled = true;
		}
		else { // in main area
			ProcessPenUpDownEvent(e);
			handled = true;
		}
		break;

	case penUpEvent:
		if (rfbsock <= 0) {
			break;
		}

		TranslateEventCoords(e);

		if (InKeyboardArea(e->screenX, e->screenY)) {
			ProcessKeyboardPenUpEvent(e);
			handled = true;
		}
		else if (InMainArea(e->screenX, e->screenY)) {
			ProcessPenUpDownEvent(e);
			handled = true;
		}
		else {
			buttonState = 0;
		}

		break;

	case penMoveEvent:
		if (rfbsock <= 0) {
			break;
		}

		TranslateEventCoords(e);

		if (buttonState == HORIZ || buttonState == VERT /*!InMainArea (e->screenX, e->screenY)*/) {
		  ProcessScrollPenMoveEvent(e);
		  handled = true;
		  break;
		}
		else {
		  if (!(InKeyboardArea(e->screenX, e->screenY))) {
				ProcessPenMoveEvent(e);
			}
		}

		handled = true;
		break;

	case keyDownEvent:
    // Check if this key is mapped, if so, run the action from the map instead.
    if (RunKeyMapAction(e->data.keyDown)) {
			handled = true;
      break;
    }

		/* Allow keyboard areas of Graffiti to activate special keys menu */
		if  ((e->data.keyDown.modifiers & commandKeyMask) &&
			(/* (e->data.keyDown.chr == vchrKeyboard) || */
			 (e->data.keyDown.chr == vchrKeyboardNumeric) /* ||
			 (e->data.keyDown.chr == vchrKeyboardAlpha) */ ))
		{
			if (rfbsock <= 0) {
				break;
			}
			ToggleKeyboardState();
			handled = true;
			KeySetMask(~KeyCurrentState());
			UpdateViewport();
			break;
		}

		// the hardware key 1 (DateBook) used to emulate the middle mouse button
		if (e->data.keyDown.chr == vchrHard1) {
			if (currentMouseButton) {
				if (KeyCurrentState() & keyBitHard2) {
					currentMouseButton = 0;
				}
				else {
					currentMouseButton = 2;
				}
			}
			handled = true;
			// Mask off the key to avoid repeat keys causing clicking sounds
			KeySetMask(~KeyCurrentState());
			break;
		}
		
		// the hardware key 2 (PhoneBook) used to emulate the right mouse button
		if (e->data.keyDown.chr == vchrHard2) {
			if (currentMouseButton) {
				if (KeyCurrentState() & keyBitHard1) {
					currentMouseButton = 0;
				}
				else {
					currentMouseButton = 4;
				}
			}
			handled = true;
			// Mask off the key to avoid repeat keys causing clicking sounds
			KeySetMask(~KeyCurrentState());
			break;
		}

		if (ProcessScrollKeyEvent(e))	{
			PositionViewport();
			PositionScrollbars();
			handled = true;
			break;
		}

		if ((!(e->data.keyDown.modifiers & commandKeyMask)) 
				&& TxtGlueCharIsValid(e->data.keyDown.chr))
		{
			SendKeySequence(0, 0, 0, &(e->data.keyDown.chr), 1, false);
		}

		handled = true;
		break;
		
		
	case winDisplayChangedEvent: 
	case displayExtentChangedEvent:
	{
		Coord width, height;
		WinGetDisplayExtent(&width, &height);
		width *= density;
		height *= density;

		if (width != screenWidth || height != screenHeight) {
			// screen size has changed (probably due to silkscreen display/hide)
			FormPtr frm = FrmGetActiveForm();

			screenWidth = width;
			screenHeight = height;
			
			WinDeleteWindow(offScreenWindow, false);
			offScreenWindow = NULL;
			CreateWindows(frm);

			SendFramebufferUpdateRequest(
				(int) palmFB.virtual.x, (int) palmFB.virtual.y,
				(int) palmFB.virtual.w,	(int) palmFB.virtual.h,
				false);
			}
		}
		break;
	

	default:
		break;
	}

	return handled;
}


/******************************************************************
 * ApplicationHandleEvent
 *
 * loads the form and sets up event handlers
 *****************************************************************/
static Boolean AppHandleEvent(EventPtr e)
{
	FormPtr frm;
	UInt16    formId;
	Boolean handled = false;

	if (e->eType == frmLoadEvent) {
		formId = e->data.frmLoad.formID;
		frm = FrmInitForm(formId);
		FrmSetActiveForm(frm);

		switch(formId) {
			case MainForm:
				FrmSetEventHandler(frm, MainFormHandleEvent);
				break;

			case ConsoleForm:
				FrmSetEventHandler(frm, ConsoleFormHandleEvent);
				break;

//			case ServerSelectForm:
//				FrmSetEventHandler(frm, ServerSelectFormHandleEvent);
//				break;

//			case ServerEditForm:
//				FrmSetEventHandler(frm, ServerEditFormHandleEvent);
//				break;

//			case ServerDetailsForm:
//				FrmSetEventHandler(frm, ServerDetailsFormHandleEvent);
//				break;

//			case PasswordForm:
//				FrmSetEventHandler(frm, PasswordFormHandleEvent);
//				break;
				
//			case PrefsForm:
//				FrmSetEventHandler(frm, PrefsFormHandleEvent);
//				break;

//			case MacroListForm:
//				FrmSetEventHandler(frm, MacroListHandleEvent);
//				break;

		}
		handled = true;
	}
	
	return handled;
}



/******************************************************************
 * StartApplication
 *
 * Get preferences, open (or create) app database
 *****************************************************************/
static UInt16 AppStart(void) {
  Err err;
	MemHandle h;
  RGBColorType* col;
  UInt16 prefsSize;
  int i;

  // Init our terminal functions for printing to the screen
  // They allow to use printf() & Co. with the output to
  // a text field in the Console form
  StdInit(ConsoleForm, ConsoleConsoleIOField, ConsoleConsoleScrollBar);

  // Get the refNum of the Net Library
  err = SysLibFind("Net.lib", &AppNetRefnum);
  if (err) {
		ErrDisplay("\nNet Library not found, exiting...");
		return err;
  }
  
  // Read the saved preferences / saved-state information
  prefsSize = sizeof(VncPreferenceType);
  if (PrefGetAppPreferences(CREATOR_ID, 0, &VNC_prefs, &prefsSize, true) 
  	== noPreferenceFound)
  {
  	// no prefs found, set default values
  	VNC_prefs.scrollbarWidth = 8;
  	VNC_prefs.disableProgressBar = true;
  	VNC_prefs.scrollBgColor = colorMineralGray;
  	VNC_prefs.scrollSliderColor = colorBlack;
  	VNC_prefs.screenPanningProp = PANNING_PROP_32;
  }


	//
	// init graphic mode
	//	
	
	// Try to set the best screen mode
	do {
	  UInt32 version;
		SonySysFtrSysInfoP sonySysFtrSysInfoP;

		/* First check for Palm OS 5 high-density API */
		err = FtrGet(sysFtrCreator, sysFtrNumWinVersion, &version);
		if (!err && version >= 4) {
			UInt32 attr;
			WinScreenGetAttribute(winScreenDensity, &attr);
			if (attr == kDensityDouble) {
				// OS5 320x320
				screenWidth = 320;
				screenHeight = 320;
				screenDepth = 8;
				//screenDepth = 16;
				density = 2;
				err = ScrDisplayMode(scrDisplayModeSet,
							&screenWidth, &screenHeight, &screenDepth, NULL);
				if (! err) {
					VNC_screenMode = modeHiDensity;
					break;
				}
			}
			else if (attr == kDensityLow) {
				// OS5 160x160 (Treo 600)
				screenWidth = 160;
				screenHeight = 160;
				screenDepth = 8;
				//screenDepth = 16;
				density = 1;
				err = ScrDisplayMode(scrDisplayModeSet,
							&screenWidth, &screenHeight, &screenDepth, NULL);
				if (! err) {
					VNC_screenMode = modeHiDensity;
					break;
				}
			}
		}
		
		// Try to load the Sony hi-res library
		err = FtrGet(sonySysFtrCreator, sonySysFtrNumSysInfoP, (UInt32 *) &sonySysFtrSysInfoP);
		if (!err) {
			if (sonySysFtrSysInfoP->libr & sonySysFtrSysInfoLibrHR) {
				/* Sony HR available */
				err = SysLibFind(sonySysLibNameHR, &sonyHiResRefNum);
				if (err == sysErrLibNotFound) {
					err = SysLibLoad(sysFileTLibrary, sonySysFileCHRLib, &sonyHiResRefNum);
				}
				HROpen(sonyHiResRefNum);
				if (!err) {
//					Coord width, height;
//					HRWinGetDisplayExtent(sonyHiResRefNum, &width, &height);

					// Try Sony 8-bit
					screenWidth = 320;
					screenHeight = 320;
					screenDepth = 8;
					density = 2;
					err = HRWinScreenMode(sonyHiResRefNum, winScreenModeSet, 
						&screenWidth, &screenHeight, &screenDepth, NULL);
					if (! err) {
						VNC_screenMode = modeSonyHires;
						break;
					}

					// Try Sony 4-bit depth (HiRes greyscale: T415, SL10 & SJ20 )
					screenWidth = 320;
					screenHeight = 320;
					screenDepth = 4;
					density = 2;
					err = HRWinScreenMode(sonyHiResRefNum, winScreenModeSet, 
						&screenWidth, &screenHeight, &screenDepth, NULL);
					if (! err) {
						VNC_screenMode = modeSonyHires;
						break;
					}
				}
			}
		}
		
		// try HandEra 330
		if (FtrGet(TRGSysFtrID, TRGVgaFtrNum, &version) == 0) {
			if (sysGetROMVerMajor(version) >= 1) {
				err = VgaSetScreenMode(screenMode1To1, rotateModeNone);
				if (! err) {
					Coord width, height;
					WinGetDisplayExtent(&width, &height); 
					screenWidth = width;
					screenHeight = height;
					screenDepth = 4;
					WinScreenMode(winScreenModeSet, NULL, NULL, &screenDepth, NULL);
					density = 1;
					VNC_screenMode = modeHandEra330;
					break;
				}
			}
		}		
		
		
		// Try legacy 8-bit colour
		screenWidth = 160;
		screenHeight = 160;
		screenDepth = 8;
		density = 1;
		err = ScrDisplayMode(scrDisplayModeSet,
					&screenWidth, &screenHeight, &screenDepth, NULL);
		if (! err) {
			VNC_screenMode = modeLegacy8bpp;
			break;
		}
					
		// No luck, try 4-bit greyscale
		screenDepth = 4;
		err = ScrDisplayMode(scrDisplayModeSet,
					&screenWidth, &screenHeight, &screenDepth, NULL);
		if (! err) {
			VNC_screenMode = modeLegacy4bpp;
			break;
		}
		
		// No luck, let's try dropping back to 2-bit mode
		screenDepth = 2;
		err = ScrDisplayMode(scrDisplayModeSet,
					&screenWidth, &screenHeight, &screenDepth, NULL);
		if (! err) {
			VNC_screenMode = modeLegacy2bpp;
			break;
		}
		
		// failed to set suitable display mode
		FrmAlert(ScreenModeAlert);
		return err;
		
	} while (false);


	if (screenDepth == 8) {
		// palette mode, init palette
		col = malloc(sizeof(RGBColorType) * 256);
		// Set color palette to match BGR233
		for (i = 0; i < 256; i++) {
			col[i].index = 0;
			col[i].r = (((i >> 0) & 0x7) << 5) | (((i >> 0) & 0x7) << 2) | (((i >> 0) & 0x7) >> 1);
			col[i].g = (((i >> 3) & 0x7) << 5) | (((i >> 3) & 0x7) << 2) | (((i >> 3) & 0x7) >> 1);
			col[i].b = (((i >> 6) & 0x3) << 6) | (((i >> 6) & 0x3) << 4) | (((i >> 6) & 0x3) << 2) | ((i >> 6) & 0x3);
		}
		WinPalette(winPaletteSet, 0, 256, col);
		free(col);
	}

	// Collapsible area initialization
	// Handles Sony and platform APIs
	CollapseAppStart();


  // open databases
  if ((serversDbRef = OpenDatabase(&serversDbInfo)) == 0) {
		return dmErrCantOpen;
	}
  if ((macrosDbRef = MacroOpenDatabase()) == 0) {
		return dmErrCantOpen;
	}
  if ((KeyMapOpenDatabase()) == 0) {
		return dmErrCantOpen;
	}

	// retrieve version
	h = DmGetResource(verRsc, 1000);
	StrPrintF(palmvncVersion, "%s", (Char *) MemHandleLock(h));
	MemHandleUnlock(h);
	DmReleaseResource(h);
	
	// Turn off jog assist if its ok
	DisableJogAssist();

  FrmGotoForm(MainForm);
  return 0;
}


/******************************************************************
 * StopApplication
 *
 * Save preferences, close forms, close app database
 *****************************************************************/
static void AppStop(void)
{
	CollapseAppStop();

	FrmSaveAllForms();
  FrmCloseAllForms();

	ShutdownNetwork(false);

	StdFree();

	CloseDatabase(serversDbRef, &serversDbInfo);
	MacroCloseDatabase(macrosDbRef);
	KeyMapCloseDatabase(keyMapDbRef);

	if (screenDepth == 8) {
		WinPalette(winPaletteSetToDefault, 0, 0, 0);
	}

	ScrDisplayMode(scrDisplayModeSetToDefaults,
				NULL, NULL, NULL, NULL);

	if (sonyHiResRefNum != sysInvalidRefNum) {
		HRClose(sonyHiResRefNum);
	}
}


/******************************************************************
 * EventLoop
 *
 * The main event loop
 *****************************************************************/
static void AppEventLoop(void)
{
  Err err;
  EventType e;
  fd_set fds;
  long eventTimeout;
  Boolean passSystem = false;

  do {
  
		if (rfbsock)
		  eventTimeout = 0; // we sleep in select
		else
		  eventTimeout = evtWaitForever;

		EvtGetEvent(&e, eventTimeout);


		if (e.eType == winExitEvent) {
			if (e.data.winExit.exitWindow == (WinHandle) FrmGetFormPtr(MainForm)) {
				dontDraw = true;
			}
		}
		else if (e.eType == winEnterEvent) {
			if (e.data.winEnter.enterWindow == (WinHandle) FrmGetFormPtr(MainForm) &&
				e.data.winEnter.enterWindow ==	(WinHandle) FrmGetFirstForm ())
			{
				dontDraw = false;
			}
		}

		// we capture the DateBook and PhoneBook keys to emulate
		// the second and the third mouse buttons
		passSystem = e.eType == keyDownEvent &&
			(e.data.keyDown.chr == hard1Chr ||
			 e.data.keyDown.chr == hard2Chr ||
			 e.data.keyDown.chr == hard3Chr ||
			 e.data.keyDown.chr == hard4Chr ||
			 e.data.keyDown.chr == pageUpChr ||
			 e.data.keyDown.chr == pageDownChr ||
			 /* e.data.keyDown.chr == vchrKeyboard ||
			 e.data.keyDown.chr == vchrKeyboardAlpha || */
			 e.data.keyDown.chr == vchrKeyboardNumeric) ;

		if (passSystem) {
		  if (!AppHandleEvent(&e)) {
				FrmDispatchEvent(&e);
			}
		}
		else {
		  if (! SysHandleEvent(&e)) {
				if (! MenuHandleEvent(NULL, &e, &err)) {
				  if (! AppHandleEvent(&e)) {
						FrmDispatchEvent(&e);
					}
				}
			}
		}

		if (rfbsock <= 0) {
		  continue;
		}

		FD_ZERO(&fds);
		FD_SET(sysFileDescStdIn, &fds);
		FD_SET(rfbsock,&fds);

		if (updateViewportSize) {

		  WinHandle prevDrawWin = WinGetDrawWindow();
		  WinSetDrawWindow(offScreenWindow);
		  WinEraseWindow();
		  WinSetDrawWindow(prevDrawWin);

		  viewportX = palmFB.virtual.x;
		  viewportY = palmFB.virtual.y;

			/* set up the size of scrollbar slides */
			horizSlideRect.extent.x =
			(Int16) (((UInt32) palmFB.virtual.w * (UInt32) palmFB.virtual.w) /
				palmFB.remote.w);
			vertSlideRect.extent.y =
			(Int16) (((UInt32) palmFB.virtual.h * (UInt32) palmFB.virtual.h) /
				palmFB.remote.h);

			/* set the pen time tracing, used to catch double-taps */
			lastPenTime = TimGetTicks();

			SendFramebufferUpdateRequest(
				palmFB.virtual.x, palmFB.virtual.y,
				palmFB.virtual.w, palmFB.virtual.h,
				false);

			UpdateViewport();
			PositionScrollbars();
			updateViewportSize = false;
			sendUpdateRequest = false;
		}

		if (sendUpdateRequest) {
		  if (!SendIncrementalFramebufferUpdateRequest()) {
				ShutdownConnection();
			}
		  sendUpdateRequest = false;
		}

	// the BSD select() call cannot be compiled with GCC, use the native one
	//   if (select (FD_SETSIZE, &fds, NULL, NULL, NULL) < 0) {

		if (NetLibSelect(AppNetRefnum, rfbsock+1, &fds, NULL, NULL,
				SelectNetTimeout, &err) < 0)
		{
		  printf("NetLibSelect returned error\n");
		  ShutdownConnection();
		}

		if (FD_ISSET(rfbsock, &fds)) {
		  if (!HandleRFBServerMessage()) {
				ShutdownConnection();
			/* FrmPopupForm(ConsoleForm); */
				PostProgressMessage("Connection Lost!", 0, 0);
		  }
//		  UpdateViewport();
		}
	#ifdef DEBUG_FLOW
		  else if (!FD_ISSET(sysFileDescStdIn, &fds))
				printf("Select returned nothing, timeout?\n");
	#endif

  } while (e.eType != appStopEvent);
}



/******************************************************************
 * PilotMain
 *
 *
 *****************************************************************/
UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags) {
  Err err;

  if (cmd == sysAppLaunchCmdNormalLaunch) {
		err = RomVersionCompatible(PALMVNC_MIN_OS, launchFlags);
		if (err) {
			return err;
		}

		err = AppStart();
		if (err) {
			return err;
		}

		sendUpdateRequest = true;
		AppEventLoop();

		AppStop();
		
  } else {
		return sysErrParamErr;
  }

  return 0;
}
