/*
 *  PalmVNC Version 2.0 - A HiRes VNC client for the Palm OS
 *
 *  Copyright (C) 2003 Olivier Grardin (olivier@palmattitude.org)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 *
 */

#include <PalmOS.h>
#include <vga.h>

#include "CollapseUtils.h"
#include "Util.h"

#include "callback.h"

#include "palmVNC.h"
#include "palmVNC_res.h"

#include "ServerEdit.h"

#include "ServerSelect.h"



DmOpenRef serversDbRef;
ServersDbHeader serversDbInfo;



/******************************************************************
 * OpenDatabase
 *
 * Opens the database, or converts database from previous version, and load preferences
 *****************************************************************/
DmOpenRef OpenDatabase(ServersDbHeader* serversDbInfo) {
	UInt index = 0;
	VoidHand recHandle;
	VoidPtr recPointer;
	DmOpenRef db;
	UInt16 err, oldRecordSize;
	Bool create_new = false;
	DmSearchStateType search_data;
	UInt16 cardNo;
	LocalID dbID;

	// First, check that if the database does exist, it is the correct version
	if (!(db = DmOpenDatabaseByTypeCreator(SERVERS_DB_TYPE, CREATOR_ID, dmModeReadWrite))) {
		create_new = true;
	}
	else {
		recHandle = DmGetRecord(db, 0);
		recPointer = MemHandleLock(recHandle);
		oldRecordSize = MemPtrSize(recPointer);
		if (oldRecordSize != sizeof(ServersDbHeader)) {
			create_new = true;
		}
		else {
			MemSet(serversDbInfo, sizeof(ServersDbHeader), (Byte) 0);
			MemMove(serversDbInfo, recPointer, sizeof(ServersDbHeader));
			if (serversDbInfo->databaseVersion != SERVERS_DB_VERSION) {
				create_new = true;
			}
		}
		MemHandleUnlock(recHandle);
		DmReleaseRecord(db, 0, true);
		DmCloseDatabase(db);
	}


	if (create_new == true) {
		// Create new database
		printf("Creating new database.\n");
		err = DmGetNextDatabaseByTypeCreator(true, &search_data, SERVERS_DB_TYPE, CREATOR_ID, false, &cardNo, &dbID);
		while (err == errNone) {
			DmDeleteDatabase(cardNo, dbID);
			err = DmGetNextDatabaseByTypeCreator(false, &search_data, SERVERS_DB_TYPE, CREATOR_ID, false, &cardNo, &dbID);
		}

		if ((err = DmCreateDatabase(0, SERVERS_DB_NAME, CREATOR_ID, SERVERS_DB_TYPE, false))) {
			return 0;
		}

		serversDbInfo->databaseVersion = SERVERS_DB_VERSION;
		serversDbInfo->lastServer = 0;

		db = DmOpenDatabaseByTypeCreator(SERVERS_DB_TYPE, CREATOR_ID, dmModeReadWrite);
		recHandle = DmNewRecord(db, &index, sizeof(ServersDbHeader));
		DmWrite(MemHandleLock(recHandle), 0, serversDbInfo, sizeof(ServersDbHeader));
		MemHandleUnlock(recHandle);
		DmReleaseRecord(db, index, true);
	}
	else {
		// Load an existing database
		db = DmOpenDatabaseByTypeCreator(SERVERS_DB_TYPE, CREATOR_ID, dmModeReadWrite);
		recHandle = DmGetRecord(db, 0);
		recPointer = MemHandleLock(recHandle);
		MemSet(serversDbInfo, sizeof(ServersDbHeader), (Byte) 0);
		MemMove(serversDbInfo, recPointer, sizeof(ServersDbHeader));
		MemHandleUnlock(recHandle);
		DmReleaseRecord(db, 0, true);
	}

  return db;
}

/******************************************************************
 * CloseDatabase
 *
 * Save preferences information.
 *****************************************************************/
void CloseDatabase(DmOpenRef db, ServersDbHeader* serversDbInfo) {
  VoidPtr p = MemHandleLock(DmGetRecord(db, 0));
  DmWrite(p, 0, serversDbInfo, sizeof(ServersDbHeader));
  MemPtrUnlock(p);
  DmReleaseRecord(db, 0, true);
  DmCloseDatabase(db);
}




/******************************************************************
 * ServerSelect_PopulateList
 *
 * Fills in the list of available serversDbInfo
 *****************************************************************/
static Boolean ServerSelect_PopulateList(FormPtr frm, Boolean do_populate) {
	ListPtr listPtr;
	static int servers_list_count = 0;
	static MemHandle servers_listH = NULL;
	static MemHandle serverH_H = NULL;
	static MemHandle* serverH = NULL;
	static Char** servers_list = NULL;

	UInt16 NumServers;
	VoidHand recHandle;
	ServersDbItem* recPointer;
	UInt16 index;

	/* Clear any old lists */
	listPtr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ServerSelectServerSelectList));
	LstSetListChoices(listPtr, NULL, 0);
	
	if (servers_list_count != 0) {
		for (index = 0; index < servers_list_count; index++) {
			MemHandleUnlock(serverH[index]);
			MemHandleFree(serverH[index]);
		}
		MemHandleUnlock(serverH_H);
		MemHandleFree(serverH_H);
		MemHandleUnlock(servers_listH);
		MemHandleFree(servers_listH);
		servers_list_count = 0;
	}
	
	if (do_populate == false) {
		return false;
	}
	
	/* Look for stored currentServer configurations */
	NumServers = DmNumRecordsInCategory(serversDbRef, dmAllCategories) - 1;
	if (NumServers == 0) {
		return false;
	}

	/* Build an index of available servers */
	servers_list_count = NumServers;
	servers_listH = MemHandleNew(sizeof(char*) * servers_list_count);
	servers_list = MemHandleLock(servers_listH);
	serverH_H = MemHandleNew(sizeof(MemHandle) * servers_list_count);
	serverH = MemHandleLock(serverH_H);
	index = 1;
	while ((recHandle = DmQueryNextInCategory(serversDbRef, &index, dmAllCategories)) != NULL) {
		recPointer = MemHandleLock(recHandle);
		serverH[index-1] = MemHandleNew(sizeof(char) * (StrLen(recPointer->name) + 1));
		servers_list[index-1] = MemHandleLock(serverH[index-1]);
		MemMove(servers_list[index-1], recPointer->name, sizeof(char) * (StrLen(recPointer->name) + 1));
		MemHandleUnlock(recHandle);
		index++;
	}
	LstSetListChoices(listPtr, servers_list, servers_list_count);
	if ((serversDbInfo.lastServer > 0) && (serversDbInfo.lastServer <= servers_list_count)) {
		LstSetSelection(listPtr, serversDbInfo.lastServer - 1);
	}
	else {
		LstSetSelection(listPtr, 0);
	}
		
	return true;
}


/******************************************************************
 * LoadServerConnection
 *
 * Copies the server connection settings from the appropriate database record
 *****************************************************************/
static void LoadServerConnection(DmOpenRef db, ServersDbItem *currentServer, UInt16 index)
{
	VoidHand recHandle;
	VoidPtr recPointer;

	char *dummy_name = "My currentServer\0";
	char *dummy_addr = "host.company.com\0";
	char *dummy_username = "username\0";
	char *dummy_passwd = "\0";

	Bool loaded = false;

	if (index > 0) {
		// Load an exisiting profile.
		recHandle = DmGetRecord(db, index);
		if (recHandle != NULL) {
			recPointer = MemHandleLock(recHandle);
			if (recPointer != NULL) {
				MemSet(currentServer, sizeof(ServersDbItem), (Byte)0);
				MemMove(currentServer, recPointer, sizeof(ServersDbItem));
				loaded = true;
				currentServer->index = index;
				MemHandleUnlock(recHandle);
			}
			DmReleaseRecord(db, index, true);
		}
	}

	if (loaded == false) {
		// fill default values
		currentServer->index = 0;
		MemMove(currentServer->name, dummy_name, StrLen(dummy_name)+1);
		MemMove(currentServer->address, dummy_addr, StrLen(dummy_addr)+1);
		MemMove(currentServer->password, dummy_passwd, StrLen(dummy_passwd)+1);
		currentServer->basePort = DEFAULT_VNC_DATA_PORT;
		currentServer->displayNumber = 0;
		currentServer->beginX = 0;
		currentServer->beginY = 0;
		currentServer->scaleFactor = 1;
		currentServer->sharedDesktop = true;
		MemMove(currentServer->loginNameNT, dummy_username, StrLen(dummy_username)+1);
		MemMove(currentServer->passwNT, dummy_passwd, StrLen(dummy_passwd)+1);
	}

	return;
}

/******************************************************************
 * SaveServerConnection
 *
 *  Copies the currentServer connection settings to the database
 *****************************************************************/
static UInt16 SaveServerConnection(DmOpenRef db, ServersDbItem *currentServer)
{
	VoidHand recHandle;
	VoidPtr recPointer;

	UInt16 local_index;

	local_index = currentServer->index;

	if (local_index > 0) {
		// Overwrite an existing profile.
		recHandle = DmGetRecord(db, currentServer->index);
		if (recHandle != NULL) {
			recPointer = MemHandleLock(recHandle);
			if (recPointer == NULL) {
				local_index = 0;
			}
			else {
				MemHandleUnlock(recHandle);
			}
			DmReleaseRecord(db, currentServer->index, true);
		}
		else {
			local_index = 0;
		}
	}

	if (local_index == 0) {
		// Create a new profile.
		local_index = dmMaxRecordIndex;
		recHandle = DmNewRecord(db, &local_index, sizeof(ServersDbItem));
		recPointer = MemHandleLock(recHandle);
		DmWrite(recPointer, 0, currentServer, sizeof(ServersDbItem));
		MemHandleUnlock(recHandle);
		DmReleaseRecord(db, local_index, true);
	}
	else {
		recHandle = DmGetRecord(db, local_index);
		recPointer = MemHandleLock(recHandle);
		DmWrite(recPointer, 0, currentServer, sizeof(ServersDbItem));
		MemHandleUnlock(recHandle);
		DmReleaseRecord(db, local_index, true);
	}
	currentServer->index = local_index;

	return local_index;
}


/******************************************************************
 * DeleteServerConnection
 *
 *  Removes currentServer connection settings from the database
 *****************************************************************/
static void DeleteServerConnection(DmOpenRef db, UInt16 index)
{
	UInt16 local_index;

	local_index = index;

	if (local_index > 0) {
		// Delete an existing profile.
		DmRemoveRecord(db, index);
	}
}



/******************************************************************
 * ServerSelectFormHandleEvent
 *
 * Handles events in the currentServer selection form
 *****************************************************************/
static Boolean ServerSelectFormHandleEvent(EventPtr e) {
	Boolean handled, closeServerSelectForm;
	FormPtr frm;
	UInt16 index;
	ListPtr listPtr;

	CALLBACK_PROLOGUE

	frm = FrmGetActiveForm();
	handled = false;
	closeServerSelectForm = false;

	switch (e->eType) {
		case frmCloseEvent:
			closeServerSelectForm = true;
			handled = true;
			break;

		case ctlSelectEvent:
			switch(e->data.ctlSelect.controlID) {

				case ServerSelectNewButton:
				case ServerSelectEditButton:
					if (e->data.ctlSelect.controlID == ServerSelectNewButton) {
						index = 0;
					}
					else {
						listPtr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ServerSelectServerSelectList));
						index = LstGetSelection(listPtr) + 1;
					}
					LoadServerConnection(serversDbRef, &currentServer, index);
					if (DoServerEdit()) {
						SaveServerConnection(serversDbRef, &currentServer);
					}
					else {
						LoadServerConnection(serversDbRef, &currentServer, index);
					}
					serversDbInfo.lastServer = index;
					ServerSelect_PopulateList(frm, true);
					FrmDrawForm(frm);
					handled = true;
					break;

				case ServerSelectCopyButton:
				case ServerSelectDeleteButton:
					listPtr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ServerSelectServerSelectList));

					switch (e->data.ctlSelect.controlID) {
						case ServerSelectCopyButton:
							index = LstGetSelection(listPtr);
							if (index != noListSelection) {
								index++;
								LoadServerConnection(serversDbRef, &currentServer, index);
								currentServer.index = 0;
								serversDbInfo.lastServer = SaveServerConnection(serversDbRef, &currentServer);
								ServerSelect_PopulateList(frm, true);
								FrmDrawForm(frm);
							}
							handled = true;
							break;

						case ServerSelectDeleteButton:
							index = LstGetSelection(listPtr);
							if (index != noListSelection) {
								index++;
								LoadServerConnection(serversDbRef, &currentServer, index);
								if (FrmCustomAlert(DeleteConfirmAlert, currentServer.name, "", "") == 0) {
									DeleteServerConnection(serversDbRef, index);
									serversDbInfo.lastServer = 1;
									ServerSelect_PopulateList(frm, true);
									FrmDrawForm(frm);
								}
							}
							handled = true;
							break;
					}
					break;

				default:
					 break;
			}
			break;

		default:
			break;
	}

	if (closeServerSelectForm) {
		// the memory under field text handles is freed by the UI manager
		CALLBACK_EPILOGUE
		FrmReturnToForm(MainForm);
	}

	CALLBACK_EPILOGUE
	return handled;
}


/******************************************************************
 * ServerSelect
 *
 * Generates the ServerSelect dialog
 *****************************************************************/
Boolean DoServerSelect() {
  ListPtr listPtr;
	FormPtr previousForm = FrmGetActiveForm();
	FormPtr frm = FrmInitForm(ServerSelectForm);
	UInt16 hitButton;
	UInt16 index;
	
	CollapseSetState(frm, collapseStateUser);
	SizeForm(frm);
	if (VNC_screenMode == modeHandEra330) {
		VgaFormModify(frm, vgaFormModify160To240);
	}

	FrmSetActiveForm(frm);
	FrmSetEventHandler(frm, ServerSelectFormHandleEvent);

	/* Populate list */
	if (! ServerSelect_PopulateList(frm, true)) {
		/* Request new default currentServer */
		LoadServerConnection(serversDbRef, &currentServer, 0);
		if (! DoServerEdit()) {
			if (previousForm) {
				FrmSetActiveForm(previousForm);
			}
			FrmDeleteForm(frm);
			return false;
		}
		SaveServerConnection(serversDbRef, &currentServer);
		ServerSelect_PopulateList(frm, true);
	}

	hitButton = FrmDoDialog(frm);
	
	if (hitButton == ServerSelectConnectButton) {
		listPtr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ServerSelectServerSelectList));
		index = LstGetSelection(listPtr);
		if (index != noListSelection) {
			index++;
			LoadServerConnection(serversDbRef, &currentServer, index);
			serversDbInfo.lastServer = index;
		}
		else {
			hitButton = ServerSelectCancelButton;
		}
	}

	/* Empty list */
	ServerSelect_PopulateList(frm, false);
	
	if (previousForm) {
		FrmSetActiveForm(previousForm);
	}
	FrmDeleteForm(frm);

	return (hitButton == ServerSelectConnectButton);
}


