/*
 * Util.c
 *
 * Misc utility functions for the PalmOS
 *
 *  Copyright (C) 2002 Olivier Grardin (olivier@palmattitude.org)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 *
 */
 
#include <PalmOS.h>

//#include "ColorPicker.h"

#include "Util.h"

/*********************************************************************
 * Entry Points
 *********************************************************************/

/*********************************************************************
 * Global variables
 *********************************************************************/

/*********************************************************************
 * Internal Constants
 *********************************************************************/

/*********************************************************************
 * Internal Variables
 *********************************************************************/

/*********************************************************************
 * Internal Functions
 *********************************************************************/


/*
 * FUNCTION:
 *
 * DESCRIPTION:
 *
 * PARAMETERS:
 */
void SetCheckboxSelected(FormType *frmP, UInt16 checkboxID, Boolean selected) {
	ControlType *checkbox = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, checkboxID));
	
	CtlSetValue(checkbox, selected);
}

/*
 * FUNCTION:
 *
 * DESCRIPTION:
 *
 * PARAMETERS:
 */
Boolean IsCheckboxSelected(FormType *frmP, UInt16 checkboxID) {
	ControlType *checkbox = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, checkboxID));
	
	return CtlGetValue(checkbox);
}



/*
 * FUNCTION: GetObjectPtr
 *
 * DESCRIPTION:
 *
 * This routine returns a pointer to an object in the specified form.
 *
 * PARAMETERS:
 *
 * formId
 *     id of the form to display
 *
 * RETURNED:
 *     address of object as a void pointer
 */

void * GetObjectPtr(FormType *frmP, UInt16 objectID)
{
    return FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, objectID));
}



/*
 * DESCRIPTION: This routine sets the label of a control
 *
 * PARAMETERS:
 *
 * frm
 *     pointer to the CrewMemberForm form.
 */
void ControlSetLabel(FormType *frmP, UInt16 ctlID, Char *label)
{
	ControlType *control;
	UInt16 idx;

	idx = FrmGetObjectIndex(frmP, ctlID);
	control = (ControlType *) FrmGetObjectPtr(frmP, idx);

	CtlSetLabel(control, label);
}



/*
 * DESCRIPTION: This routine sets the displayed text in a field
 *
 * PARAMETERS:
 *
 * frm
 *     pointer to the Form form.
 */
void FieldSetText(FormType *frmP, UInt16 fieldID, Char *text)
{
	FieldType *field;
	UInt16 fieldIndex;

	fieldIndex = FrmGetObjectIndex(frmP, fieldID);
	field = (FieldType *) FrmGetObjectPtr(frmP, fieldIndex);
	
	FldDelete(field, 0, FldGetTextLength(field));
	FldInsert(field, text, StrLen(text));
}




/*
 * DESCRIPTION: This routine sets the displayed text in a field
 *
 * PARAMETERS:
 *
 * frm
 *     pointer to the Form form.
 */
void FieldSetTextWithSelection(FormType *frmP, UInt16 fieldID, Char *text,
	UInt16 start, UInt16 end)
{
	FieldType *field;
	UInt16 fieldIndex;

	fieldIndex = FrmGetObjectIndex(frmP, fieldID);
	field = (FieldType *) FrmGetObjectPtr(frmP, fieldIndex);
	
	FldDelete(field, 0, FldGetTextLength(field));
	FldInsert(field, text, StrLen(text));
	FldSetSelection(field, start, end);
}




/*
 * DESCRIPTION: This routine sets the displayed field as editable or not
 *
 * PARAMETERS:
 *
 * frm
 *     pointer to the Form form.
 */
void FieldSetEditable(FormType *frmP, UInt16 fieldID, Boolean editable)
{
	FieldType *field;
	UInt16 fieldIndex;
	FieldAttrType attr;

	fieldIndex = FrmGetObjectIndex(frmP, fieldID);
	field = (FieldType *) FrmGetObjectPtr(frmP, fieldIndex);

	FldGetAttributes(field, &attr);
	attr.editable = editable;
	FldSetAttributes(field, &attr);
}




static void UtilWinDrawTruncChars(const char * str, UInt16 length,
                         Int16 x, Int16 y, Int16 maxWidth, Boolean rightJustify)
{
    Boolean fitInWidth;
    Int16 len, width;

		width = maxWidth;
    len = length;
    fitInWidth = false;		
    FntCharsInWidth(str, &width, &len, &fitInWidth);

  	if (rightJustify) {
	    WinDrawChars(str, len, x + maxWidth - width, y);
	  }
  	else {
	    WinDrawChars(str, len, x + 1, y);
  	}
  	
}

/*
 * DESCRIPTION: callback funtion that draws a date contained in the
 * table item's Int value.
 *
 * PARAMETERS:
 *
 * RETURNED:
 */

void CustomDateRenderer (void *table,
	Int16 row, Int16 column, RectangleType *bounds) {

  Char str[dateStringLength];

	Int16 dateInt = TblGetItemInt(table, row, column);
  FontID font = TblGetItemFont(table, row, column);

	FontID prevFont = FntSetFont(font);
	
	if (dateInt != -1) {
		DateType *date = (DateType*) (void*) &dateInt;

    DateToAscii(date->month, date->day, date->year + 1904,
            PrefGetPreference(prefDateFormat), str);
  }
  else {
  	StrCopy(str, "-");
  }
  
	
  WinEraseRectangle(bounds, 0);
  UtilWinDrawTruncChars(str, StrLen(str),
  	bounds->topLeft.x, bounds->topLeft.y, bounds->extent.x, true);
	
  FntSetFont(prevFont);
}

/*
 * DESCRIPTION: callback funtion that draws a string pointed to by
 * the items's Ptr value
 *
 * PARAMETERS:
 *
 * RETURNED:
 */

void CustomStringRenderer (void *table,
	Int16 row, Int16 column, RectangleType *bounds) {

  Char * str = (Char *) TblGetItemPtr(table, row, column);
  FontID font = TblGetItemFont(table, row, column);

	FontID prevFont = FntSetFont(font);
	
  WinEraseRectangle(bounds, 0);
  UtilWinDrawTruncChars(str, StrLen(str),
  	bounds->topLeft.x, bounds->topLeft.y, bounds->extent.x, false);
  	
  FntSetFont(prevFont);
}

/*
 * DESCRIPTION: callback funtion that draws an icon specified by
 * the resource ID of a bitmap taken from	the table item's Int value
 *
 * PARAMETERS:
 *
 * RETURNED:
 */
void CustomIconRenderer (void *table,
	Int16 row, Int16 column, RectangleType *bounds) {

	DmResID resID = (DmResID) TblGetItemInt(table, row, column);
	MemHandle hBitmap = DmGetResource(bitmapRsc, resID);
 
  WinEraseRectangle(bounds, 0);
	if (hBitmap) {
		BitmapPtr pBitmap = (BitmapPtr) MemHandleLock(hBitmap);
		WinDrawBitmap(pBitmap, bounds->topLeft.x, bounds->topLeft.y);
		MemHandleUnlock(hBitmap);
		DmReleaseResource(hBitmap);
	}
}

/*
 * DESCRIPTION: resize form according to actual screen size
 *
 * PARAMETERS:
 *
 * RETURNED:
 */
void SizeForm(FormPtr frmP)
{
	RectangleType rect;
	
	rect.topLeft.x = 0;
	rect.topLeft.y = 0;
	WinGetDisplayExtent(&rect.extent.x, &rect.extent.y);
	WinSetBounds(FrmGetWindowHandle(frmP), &rect);
}